# -*- coding: utf8 -*-
"""

******************
``projectflow.py``
******************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``projectflow.py`` is a module to handle the project aspects of the tool-flow. This module is responsible for
generating source code (if enabled) using ``xml2vhdl``, collating all of the source-code for the project,
and adding all ``HDL`` files to a ``VUnit`` object which can then be used to perform operations relevant
to the current project.

"""
import helpers.customlogging as log
import helpers.arguments as arguments
import helpers.version as my_version
import helpers.funcs as funcs

import projectmanager as projectmanager
import xml2vhdl_iface as xml2xml_iface
import fpgavendor_iface as fpgavendor_iface
import vunit_iface as vunit_iface

import os
import shutil

logger = log.config_logger(name=__name__)
LOGLEVEL = log.LogLevelsConsts

rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = my_version.Version(0, 2, 0, svn_rev=rev, disable_svn_logging=True)
__version__ = version.get_version()
__str__ = my_version.about(__name__, __version__, version.revision)


class ProjectFlow(projectmanager.RepoFlow):
    """

    """

    def __init__(self, **kwargs):
        super(ProjectFlow, self).__init__()
        log.sect_break(self.logger)
        self.logger.info('Processing Project...')
        log.sect_break(self.logger)

        hdl_file_ext = ['.vhd', '.vhdl', '.v']

        if self.vendor == 'xilinx':
            ip_ext = ['.xci']
            constr_ext = ['.xdc']
        elif self.vendor == 'altera':
            ip_ext = ['.ip', '.qsys', '.qip']
            constr_ext = ['.sdc', '.qsf', '.tcl']

        all_ip_files = list()
        all_ip_wrapper_files = list()
        all_constr_files = list()
        ip_wrapper_files_no_dups = list()
        exclude_hdl_keywords = ['/vhd', '/verilog']
        exclude_hdl_keywords.extend(self.exclude_keywords)

        exclude_ip_keywords = ['/ip']
        exclude_ip_keywords.extend(self.exclude_keywords)

        try:
            setattr(self, 'family_mappings', self.config['fpga'][self.vendor]['family_mappings'])
        except KeyError:
            setattr(self, 'family_mappings', dict())

        if self.xml2vhdl_support:
            self.process_xml2vhdl(dependencies=self.dependencies,
                                  settings_file=self.settings_file,
                                  clean=True)
        else:
            self.logger.info('XML2VHDL Support: Disabled')

        # \todo Check when vendor_ip locations are available for processing XML2VHDL.

        log.sect_break(self.logger)
        self.logger.info('Processing EXCLUDE Terms...')
        self.logger.info('Using Exclude Keywords from: {config_file}'.format(config_file=self.config_file))
        for kw in self.exclude_keywords:
            self.logger.info('\t{kw}'.format(kw=kw))

        supported_vendors = self.config['fpga']['supported_vendors']
        self.exclude_keywords = self.add_vendor_to_exclude_terms(exclude_terms=self.exclude_keywords,
                                                                 vendor=self.vendor,
                                                                 supported_vendors=supported_vendors)

        self.logger.info('Updated Exclude Keywords with Vendors to Exclude:')
        for kw in self.exclude_keywords:
            self.logger.info('\t{kw}'.format(kw=kw))

        vunit_hdl_dict = dict()
        has_board = False

        self.logger.info('Searching for Top-Level in Project...')
        for depend in self.dependencies:
            # self.logger.error('{depend}'.format(depend=dir(depend)))
            if depend.top_level:
                self.logger.info('Found Top-Level: {depend.lib}.{depend.name}'.format(depend=depend))
                setattr(self, 'lib', getattr(depend, 'lib', None))
                setattr(self, 'project_lib', self.lib)
                setattr(self, 'local_path', getattr(depend, 'local_path', None))
                setattr(self, 'simulation_dir', getattr(depend, 'simulation_dir', None))
                top_level_dependency = depend
                break

        self.logger.info('Searching for Board in Project...')
        for depend in self.dependencies:
            if depend.is_board:
                self.logger.info('Found Board: {depend.lib}.{depend.name}'.format(depend=depend))
                if not hasattr(self, 'local_path'):
                    setattr(self, 'lib', getattr(depend, 'lib', None))
                    setattr(self, 'project_lib', self.lib)
                    local_path = os.path.join(getattr(depend, 'local_path', None),
                                              self.vendor,
                                              depend.name,
                                              self.tool_version)
                    setattr(self, 'local_path', local_path)
                    setattr(self, 'simulation_dir', getattr(depend, 'simulation_dir', None))
                    top_level_dependency = depend
                has_board = True
                setattr(self, 'project_top_level', self.top_level)
                board_priority_list = ['vendor', 'tool_version', 'family', 'device',
                                       'name', 'lib', 'top_level', 'board_id', 'board_timestamp_enabled']

                depend.set_as_board(top_level_dependency=top_level_dependency,
                                    board_priority_list=board_priority_list)

                setattr(self, 'is_board', True)
                for key in board_priority_list:
                    setattr(self, key, getattr(depend, key, None))
                break

        if hasattr(self, 'board_id'):
            self.logger.debug('Processing Board ID for Project: {self.board_id}'.format(self=self))
            board_id = self.board_id
            for char in ['X', 'x', '"', '_', 'U', '\\']:
                if char in board_id:
                    board_id = board_id.replace(char, "")
            setattr(self, 'board_id', int(board_id, 16))
            self.logger.info('Processed Board ID: {self.board_id}'.format(self=self))

        for depend in self.dependencies:
            if depend.vunit_default and depend.name != 'vendor_ip':
                self.logger.info('.' * 80)
                self.logger.info('Processing: {depend.lib}.{depend.name}'.format(depend=depend))

                src_files_list = list()

                try:
                    src_files_list = self.get_files(root_path=depend.src_dir,
                                                    file_exts=hdl_file_ext,
                                                    exclude_terms=self.exclude_keywords,
                                                    exclude_filename='EXCLUDE',
                                                    growing_list=src_files_list)
                except AttributeError as e:
                    self.logger.critical('Could Not Process HDL for: {depend.name}'
                                         .format(depend=depend))
                    self.logger.critical('{e}'.format(e=e))

                if src_files_list:
                    vunit_hdl_dict[depend.lib] = src_files_list

                try:
                    all_constr_files = self.get_files(root_path=depend.constraints_dir,
                                                      file_exts=constr_ext,
                                                      exclude_terms=self.exclude_keywords,
                                                      exclude_filename='EXCLUDE',
                                                      growing_list=all_constr_files)
                except AttributeError as e:
                    self.logger.critical('Could Not Process Constraints for: {depend.name}'
                                         .format(depend=depend))
                    self.logger.critical('{e}'.format(e=e))

            if getattr(depend, 'has_vendor_ip', False):
                self._validate_vendor_ip_requirements()
                self.logger.info('Processing IP for: {depend.name}'.format(depend=depend))

                for vend_ip in self.dependencies:
                    if vend_ip.vunit_default and vend_ip.name == 'vendor_ip':
                        self.logger.info('Found Master Vendor IP Dependency...')
                        ip_dependencies = projectmanager.VendorIpDependency(repo_root=self.repo_root,
                                                                            master_vendor_ip=vend_ip,
                                                                            vendor=depend.vendor,
                                                                            tool_version=depend.tool_version,
                                                                            family=depend.family,
                                                                            device=depend.device,
                                                                            ip_dict=depend.ip)

                ip_cnt = 0
                for k, v in ip_dependencies.enabled_ip.items():
                    self.logger.info('-' * 80)
                    self.logger.info('Processing Enabled IP: {k}'.format(k=k))
                    self.logger.info('Searching for Vendor IP Design Files...')
                    generated_ip_exclude_keywords = list()
                    generated_hdl_exclude_keywords = list()

                    if self.vendor == 'altera':
                        generated_ip_path = os.path.join('/ip', k)
                        generated_hdl_exclude_keywords = [generated_ip_path]
                        generated_ip_exclude_keywords = [generated_ip_path]
                        self.logger.info('(Altera) Excluding Generated IP Path: {path}'
                                         .format(path=generated_ip_path))

                        generated_hdl_exclude_keywords.extend(exclude_hdl_keywords)
                        generated_ip_exclude_keywords.extend(exclude_ip_keywords)

                    elif self.vendor == 'xilinx':
                        generated_hdl_exclude_keywords = exclude_hdl_keywords
                        generated_ip_exclude_keywords = exclude_ip_keywords

                    self.logger.debug('Exclude Terms for IP Files: {excl}'
                                      .format(excl=generated_hdl_exclude_keywords))
                    self.logger.debug('Exclude Terms for IP Wrapper Files: {excl}'
                                      .format(excl=generated_ip_exclude_keywords))
                    try:
                        all_ip_files = self.get_files(root_path=v['src_dir'],
                                                      file_exts=ip_ext,
                                                      exclude_terms=generated_hdl_exclude_keywords,
                                                      exclude_filename='EXCLUDE',
                                                      growing_list=all_ip_files)
                    except AttributeError as e:
                        self.logger.critical('Could Not Process IP Files for Vendor IP: {k}'
                                             .format(k=k))
                        self.logger.critical('{e}'.format(e=e))

                    except KeyError as e:
                        self.logger.critical('Could Not Process IP Files for Vendor IP: {k}'
                                             .format(k=k))
                        self.logger.critical('{e}'.format(e=e))

                    self.logger.info('Searching for Vendor IP HDL Wrapper Files...')

                    try:
                        self.logger.debug('Searching for IP Wrapper Files in: {path}'
                                          .format(path=v['src_dir']))
                        all_ip_wrapper_files = self.get_files(root_path=v['src_dir'],
                                                              file_exts=hdl_file_ext,
                                                              exclude_terms=generated_ip_exclude_keywords,
                                                              exclude_filename='EXCLUDE',
                                                              growing_list=all_ip_wrapper_files)
                    except AttributeError as e:
                        self.logger.critical('Could Not Process HDL Wrapper Files for Vendor IP: {k}'
                                             .format(k=k))
                        self.logger.critical('{e}'.format(e=e))

                    except KeyError as e:
                        self.logger.critical('Could Not Process HDL Wrapper Files for Vendor IP: {k}'
                                             .format(k=k))
                        self.logger.critical('{e}'.format(e=e))

                    try:
                        all_constr_files = self.get_files(root_path=v['constraints_dir'],
                                                          file_exts=constr_ext,
                                                          exclude_terms=self.exclude_keywords,
                                                          exclude_filename='EXCLUDE',
                                                          growing_list=all_constr_files)
                    except AttributeError as e:
                        self.logger.critical('Could Not Process Constraints for Vendor IP: {k}'
                                             .format(k=k))
                        self.logger.critical('{e}'.format(e=e))

                    except KeyError as e:
                        self.logger.critical('Could Not Process Constraints for Vendor IP: {k}'
                                             .format(k=k))
                        self.logger.critical('{e}'.format(e=e))

                    ip_cnt += 1
                    self.logger.debug('#' * 80)
                    self.logger.debug('IP Pass: #{ip_cnt}'.format(ip_cnt=ip_cnt))
                    for ip_file in all_ip_files:
                        self.logger.debug('\t{ip_file}'.format(ip_file=ip_file))

                    for wf in all_ip_wrapper_files:
                        self.logger.debug('\t{wf}'.format(wf=wf))

                if all_ip_files:
                    setattr(self, 'has_vendor_ip', True)
                    setattr(self, 'ip_files', all_ip_files)
                else:
                    setattr(self, 'has_vendor_ip', False)
                    setattr(self, 'ip_files', list())

                if all_constr_files:
                    setattr(self, 'constr_files', all_constr_files)
                else:
                    setattr(self, 'constr_files', list())

                for wf in all_ip_wrapper_files:
                    fname = os.path.split(wf)[1]
                    if any(fname in f for f in ip_wrapper_files_no_dups):
                        if 'common_wrapper_files' not in wf:
                            for cwf in ip_wrapper_files_no_dups:
                                if os.path.split(wf)[1] in cwf:
                                    self.logger.info('Two Wrapper Files Found. Prioritising Wrapper File:')
                                    self.logger.info('\tremoving: {cwf}'.format(cwf=cwf))
                                    ip_wrapper_files_no_dups.remove(cwf)
                                    self.logger.info('\tadding:   {wf}'.format(wf=wf))
                                    ip_wrapper_files_no_dups.append(wf)
                    elif wf not in ip_wrapper_files_no_dups:
                        self.logger.info('Single Wrapper File Found:')
                        self.logger.info('\tadding:   {wf}'.format(wf=wf))
                        ip_wrapper_files_no_dups.append(wf)
                    else:
                        self.logger.debug('Duplicate Wrapper File Found:')
                        self.logger.debug('\tskipping: {wf}'.format(wf=wf))

                if ip_wrapper_files_no_dups:
                    vunit_hdl_dict[v['lib']] = ip_wrapper_files_no_dups

        self.logger.debug('Completed VUnit Source Files Additions')
        for k, v in vunit_hdl_dict.items():
            for f in v:
                self.logger.debug('{lib}.{f}'.format(lib=k, f=f))

        self.logger.info('-' * 80)
        self.logger.info('Name:              {self.name}'.format(self=self))
        self.logger.info('Library:           {self.lib}'.format(self=self))
        self.logger.info('Top-Level:         {self.top_level}'.format(self=self))

        setattr(self, 'top_level_files', self.get_top_level_files(vunit_hdl_dict=vunit_hdl_dict,
                                                                  top_level_lib=self.lib,
                                                                  top_level=self.top_level,
                                                                  top_ids=['top_', '_top'],
                                                                  settings_file=self.settings_file))

        if hasattr(self, 'project_top_level'):
            self.logger.info('Project Top-Level: {self.project_top_level}'.format(self=self))
            if isinstance(self.project_top_level, (list, )):
                project_top_name = self.remove_top_id(name=self.project_top_level[0])
            else:
                setattr(self, 'project_top_level', [self.project_top_level])
                project_top_name = self.remove_top_id(name=self.project_top_level[0])
            board_name = self.remove_top_id(name=self.name)
        else:
            setattr(self, 'project_top_level', None)
            top_level_stripped_path = os.path.split(self.top_level_files[0])[1]
            top_level_stripped_ext = os.path.splitext(top_level_stripped_path)[0]
            project_top_name = self.remove_top_id(name=top_level_stripped_ext)
            board_name = None

        project_name_override = self.project_name_override
        setattr(self, 'project_name', self.generate_project_name(top_level_name=project_top_name,
                                                                 project_name_override=project_name_override,
                                                                 board_name=board_name))

        if 'boards' in self.local_path:
            build_path = os.path.join(self.local_path, 'build',
                                      self.project_name)
        else:
            build_path = os.path.join(self.local_path, 'build',
                                      self.vendor, self.tool_version, self.project_name)

        setattr(self, 'build_path', build_path)
        self.logger.info('Build Path: {self.build_path}'
                         .format(self=self))

        self.logger.info('Creating VUnit Object for Current Library: {lib}'
                         .format(lib=self.lib))
        self.logger.debug('msim_64bit: {self.sim_mode}'.format(self=self))
        if self.sim_mode == 64:
            msim_mode = True
        else:
            msim_mode = False

        setattr(self, 'hdl_sim_libs', self.config['simulation'][self.vendor])

        clean_precompiled_libs = False
        for c in ['precompiled_sim']:
            if c in self.clean:
                self.logger.debug('Clean Option Found: "{c}". Cleaning Precompiled Simulation Libraries...'
                                  .format(c=c))
                clean_precompiled_libs = True
                break

        vu = vunit_iface.VUnitProject(simulation_path=self.simulation_dir,
                                      sim_pre_compiled_libs_path=self.sim_pre_compiled_libs_path,
                                      fpga_vendor=self.vendor,
                                      fpga_toolversion=self.tool_version,
                                      sim_version=self.sim_version,
                                      setting_file=self.settings_file,
                                      hdl_lib_file_dict=vunit_hdl_dict,
                                      args=self.args,
                                      clean_precompiled_libs=clean_precompiled_libs,
                                      hdl_sim_libs=self.hdl_sim_libs,
                                      msim_64bit=msim_mode,
                                      use_osvvm=False,
                                      use_verification_components=False,
                                      modelsim_installdir=self.modelsim_installdir,
                                      repo_root=self.repo_root)

        try:
            use_glbl = self.use_glbl
        except AttributeError:
            self.logger.warning("'use_glbl:' Missing in Settings File: {settings}"
                                .format(settings=self.settings_file))
            self.logger.warning("\t'Setting to: 'use_glbl: False'")
            use_glbl = False

        vu.setup_simulation_options(vendor=self.vendor,
                                    glbl=use_glbl,
                                    glbl_lib=self.lib)

        try:
            if self.args.gui:
                sim_init_script = os.path.join(self.simulation_dir,
                                               self.sim_version,
                                               self.sim_init_script)
                if os.path.isfile(sim_init_script):
                    vu.configure_gui(sim_init_script=sim_init_script,
                                     repo_root=self.repo_root)
                else:
                    self.logger.critical('Could Not Find Simulation Init Script: {file}'
                                         .format(file=sim_init_script))

        except AttributeError:
            pass

        if self.vendor == 'xilinx':
            vu.add_vivado_sim_dependencies(vivado_path=self.xilinx_vivado,
                                           glbl=use_glbl,
                                           glbl_lib=self.lib)
            project_name = self.project_name + '.xpr'

        elif self.vendor == 'altera':
            project_name = self.project_name + '.qsf'

        if self.family in self.family_mappings:
            self.logger.debug('Family Mapping Found...')
            family = self.family_mappings[self.family]
            setattr(self, 'family', family)

        else:
            self.logger.debug('Family Mapping Not Found...')

        constraint_files = list()

        proj_hdl_files = list()
        top_level_hdls = [hdl for hdl in vunit_hdl_dict[self.lib] if self.top_level[0] in hdl]
        for top in top_level_hdls:
            top_level_hdl_files = vu.get_compile_order(top_level_file=top,
                                                       top_level_lib=self.lib)
            if self.project_top_level:
                found_proj_top = False
                proj_top_level_hdls = [hdl for hdl in vunit_hdl_dict[self.project_lib]
                                       if self.project_top_level[0] in hdl]
                for src_file in top_level_hdl_files:
                    if proj_top_level_hdls == os.path.abspath(src_file.name):
                        found_proj_top = True
                        break
                if found_proj_top:
                    self.logger.info('Found Project Top-Level Found.')
                    for src_file_obj in top_level_hdl_files:
                        if [src_file_obj.library.name, os.path.abspath(src_file_obj.name)] \
                                not in proj_hdl_files:
                            proj_hdl_files.append([src_file_obj.library.name,
                                                   os.path.abspath(src_file_obj.name)])
                else:
                    for proj_top in proj_top_level_hdls:
                        proj_top_level_hdl_files = vu.get_compile_order(top_level_file=proj_top,
                                                                        top_level_lib=self.project_lib)

                        for src_file_obj in proj_top_level_hdl_files:
                            if [src_file_obj.library.name, os.path.abspath(src_file_obj.name)] \
                                    not in proj_hdl_files:
                                proj_hdl_files.append([src_file_obj.library.name,
                                                       os.path.abspath(src_file_obj.name)])
                        for src_file_obj in top_level_hdl_files:
                            if [src_file_obj.library.name, os.path.abspath(src_file_obj.name)] \
                                    not in proj_hdl_files:
                                proj_hdl_files.append([src_file_obj.library.name,
                                                       os.path.abspath(src_file_obj.name)])

        for src in proj_hdl_files:
            self.logger.debug('{src}'.format(src=src))

        rebuild_project = False
        for c in ['all', 'fpga']:
            if c in self.clean:
                self.logger.debug('Clean Option Found: "{c}". Rebuilding Project'.format(c=c))
                rebuild_project = True
                break

        fpgaproject = fpgavendor_iface.FpgaProject(project_name=project_name,
                                                   build_path=self.build_path,
                                                   top_level_entity=self.top_level,
                                                   family=self.family,
                                                   device=self.device,
                                                   hdl_src_files=proj_hdl_files,
                                                   ip_files=self.ip_files,
                                                   constraint_files=self.constr_files,
                                                   repo_root=self.repo_root,
                                                   rebuild_project=rebuild_project,
                                                   board_id=self.board_id,
                                                   board_timestamp_enabled=self.board_timestamp_enabled)

        if self.has_vendor_ip:
            self.generate_vendor_ip()
            vu.compile_project_ip(project_file=fpgaproject.get_project_file())

        log.sect_break(self.logger)
        self.logger.info('Processed Project')

        if not self.headless and self.open_gui == 'fpga':
            project_file = fpgaproject.get_project_file()
            if self.vendor == 'altera':
                project_file = os.path.splitext(project_file)[0] + '.qpf'
                self.logger.info('Opening Project: {project}'
                                 .format(project=project_file))
                fpgavendor_iface.run_quartus_sh(quartus_bin='quartus',
                                                gui=True,
                                                project_file=project_file)
            elif self.vendor == 'xilinx':
                self.logger.info('Opening Project: {project}'
                                 .format(project=project_file))
                fpgavendor_iface.run_vivado(vivado_bin='vivado',
                                            mode='gui',
                                            tcl_file_name=project_file)

        log.sect_break(self.logger)

    def generate_vendor_ip(self):
        """

        """
        for ip in self.ip_files:
            if self.vendor == 'altera':
                self.logger.info('Generating Vendor IP From File: {ip}'
                                 .format(ip=ip))
                ip_path, ip_file = os.path.split(ip)
                ip_name = os.path.splitext(ip_file)[0]
                output_path = os.path.join(ip_path, ip_name)
                self.logger.info('Generating IP in: {path}'.format(path=output_path))
                self.logger.info('Using FPGA Family: {self.family}'.format(self=self))
                self.logger.info('Using FPGA Device: {self.device}'.format(self=self))

                clean = False
                for c in ['all', 'ip']:
                    if c in self.clean:
                        self.logger.debug('Clean Option Found: "{c}". Cleaning IP'.format(c=c))
                        clean = True
                        break

                if not os.path.exists(output_path) or clean:
                    fpgavendor_iface.run_ip_generate(exe='qsys-generate',
                                                     ip_file=ip,
                                                     output_path=output_path,
                                                     synth=True,
                                                     synth_language='VERILOG',
                                                     family=self.family,
                                                     device=self.device,
                                                     clean=clean)
                    fpgavendor_iface.run_ip_generate(exe='qsys-generate',
                                                     ip_file=ip,
                                                     output_path=output_path,
                                                     synth=False,
                                                     sim_language='VHDL',
                                                     family=self.family,
                                                     device=self.device,
                                                     clean=False)
                else:
                    self.logger.warning('Skipping Generation. Output Path Already Exists: {path}'
                                        .format(path=output_path))
            elif self.vendor == 'xilinx':
                pass

    def get_files(self, **kwargs):
        """

        Keyword Args:
            **root_path (str): Absolute path to path to start searching from. Searching will be recursive.
            **file_exts (list of str): A list of file extensions, including preceding ``.``, to search for.
            **exclude_terms (list of str): A list of strings, which if any are found in the path, will
                result in the path being excluded from the search.
            **exclude_filename (str, optional): If a file matching this name is found in the search path it
                will be read line-by-line to provide additional terms for excluding files from the path
                where the exclude file is found. Default value: ``'EXCLUDE'``
            **growing_list (list of str, optional): A list to append found files to. This allows lists to
                grow. If not provided a new list will be generated. Default value: ``list()``


        Returns:
            list of str: List of absolute paths of files matching search criteria.

        """
        root_path = funcs.get_kwarg('root_path', kwargs, '')
        file_exts = funcs.get_kwarg('file_exts', kwargs, list())
        exclude_terms = funcs.get_kwarg('exclude_terms', kwargs, list())
        exclude_filename = funcs.get_kwarg('exclude_filename', kwargs, 'EXCLUDE')
        growing_list = funcs.get_kwarg('growing_list', kwargs, list())

        found_file_list = self.search_paths_for_files(root_path=root_path,
                                                      file_extensions=file_exts,
                                                      exclude_terms_list=exclude_terms,
                                                      exclude_filename=exclude_filename)

        if found_file_list:
            for f in found_file_list:
                if f not in growing_list:
                    growing_list.append(f)

        return growing_list

    def remove_top_id(self, **kwargs):
        """Removes ``top`` ID from string

        Removes ``_top`` or ``top_`` from name.

        Keyword Args:
            name (str): Name to remove ``top`` ID from.

        Returns:
            str: ``name`` with ``_top`` or ``top_`` removed.

        """
        name = funcs.get_kwarg('name', kwargs, '')

        if name.endswith('_top'):
            top_level_name = name[:-4]
        elif name.startswith('top_'):
            top_level_name = name[4:]
        else:
            top_level_name = name

        return top_level_name

    def _validate_vendor_ip_requirements(self):
        """Checks all attributes to process vendor_ip exist

        Returns:
            bool: ``True`` if successful. Performs a ``sys.exit()`` if unsuccessful.

        """
        try:
            vendor_requirements = True
            if isinstance(self.vendor, (str,)) and self.vendor != "":
                pass
            else:
                vendor_requirements = False
                self.logger.error('Can Not Process Vendor IP. Unsupported Vendor: {self.vendor}'
                                  .format(self=self))
            if isinstance(self.tool_version, (str,)) and self.tool_version != "":
                pass
            else:
                vendor_requirements = False
                self.logger.error('Can Not Process Vendor IP. Unsupported Tool Version: {self.tool_version}'
                                  .format(self=self))
            if isinstance(self.family, (str,)) and self.family != "":
                pass
            else:
                vendor_requirements = False
                self.logger.error('Can Not Process Vendor IP. Unsupported Family: {self.family}'
                                  .format(self=self))
            if isinstance(self.device, (str,)) and self.device != "":
                pass
            else:
                vendor_requirements = False
                self.logger.error('Can Not Process Vendor IP. Unsupported Device: {self.device}'
                                  .format(self=self))

            if not vendor_requirements:
                self.logger.error('Can Not Process Vendor IP. Requirements Not Met.')
                self.logger.error('Update Settings File and Retry: {self.settings_file}'
                                  .format(self=self))
                log.errorexit(self.logger)

        except AttributeError as e:
            self.logger.error('Missing Settings Required to Process Vendor IP')
            self.logger.error('{e}'.format(e=e))
            self.logger.error('Update Settings File and Retry: {self.settings_file}'
                              .format(self=self))
            log.errorexit(self.logger)

        return True

    def search_paths_for_files(self, **kwargs):
        """Recursive search of root path for files matching file extension and exclusion criteria.

        Keyword Args:
            **root_path (str or list of str): The root path(s) to recursively search
            **file_extensions (list of str): A list of file extensions to search. Must include '.' prefix
                to match requirements of ``os.path.splitext()[1]``.
            **exclude_terms_list (list of str): A list of exclude terms. A path containing `any` item in this
                list will be excluded from the search.
            **exclude_filename (str, optional): Any file found in the search paths matching this name will
                be parsed, line-by-line, to add path specific exclusions. Default value: ``'EXCLUDE'``

        Returns:
            list of str: A list of absolute paths of files matching criteria. Empty list if none found.

        """
        root_paths = funcs.get_kwarg('root_path', kwargs, list())
        file_extensions = funcs.get_kwarg('file_extensions', kwargs, list())
        exclude_terms_list = funcs.get_kwarg('exclude_terms_list', kwargs, list())
        exclude_filename = funcs.get_kwarg('exclude_filename', kwargs, 'EXCLUDE')

        if isinstance(root_paths, (str, )) and root_paths != "":
            root_paths = [root_paths]

        search_paths = list()
        exclude_file_list = list()

        found_file_list = list()

        for root_path in root_paths:
            self.logger.info('Searching for {exts} Files in: {path} '.format(exts=file_extensions,
                                                                             path=root_path))
            for root, subdirs, filenames in os.walk(root_path):
                self.logger.debug('Subpaths: {subdirs}'.format(subdirs=subdirs))
                if filenames:
                    if any(kw in root for kw in exclude_terms_list):
                        kw_matches = [kw for kw in exclude_terms_list if kw in root]
                        self.logger.debug("Filtering Path. Keyword Exclusion(s): "
                                          "'{kw}' in Path: {path}"
                                           .format(path=root, kw=kw_matches))
                    else:
                        search_paths.append(root)

            if search_paths:
                self.logger.info('Valid Search Paths:')
                for path in search_paths:
                    self.logger.info('\t{path}'.format(path=path))
            else:
                self.logger.warning('No Search Paths Found while Processing: {path}'
                                    .format(path=root_path))

            for path in search_paths:
                for root, subdirs, filenames in os.walk(path):
                    for f in filenames:
                        if f == exclude_filename:
                            exclude_file = os.path.join(root, f)
                            if exclude_file not in exclude_file_list:
                                exclude_file_list.append(exclude_file)
                                self.logger.info("'{exclude_filename}' File Found: {exclude_file}"
                                                 .format(exclude_filename=exclude_filename,
                                                         exclude_file=exclude_file))
                        else:
                            if os.path.splitext(f)[1] in file_extensions:
                                found_file_list.append(os.path.join(root, f))
                            else:
                                self.logger.debug("Filtering File. Ext Exclusion(s): "
                                                  "'{ext}' Not in {exts} for Filename: {path}"
                                                  .format(ext=os.path.splitext(f)[1],
                                                          exts=file_extensions,
                                                          path=os.path.join(root, f)))
                    break

        if found_file_list:
            self.logger.debug('Unfiltered Files:')
            for f in found_file_list:
                self.logger.debug('\t{f}'.format(f=f))
        else:
            self.logger.debug('No Files Found while Processing:')
            for path in root_paths:
                self.logger.debug('\t{path}:'
                                  .format(path=path))

        exclude_file_keywords = list()
        no_dups_file_list = list()

        if exclude_file_list:
            for f in exclude_file_list:
                self.logger.info("Parsing '{exclude_filename}' File: {exclude_file}"
                                 .format(exclude_filename=exclude_filename,
                                         exclude_file=f))
                with open(f) as fp:
                    for l in fp.read().splitlines():
                        if len(l.split()) == 0:
                            self.logger.debug("\tSkipping Empty Line in '{exclude_filename}' File"
                                              .format(exclude_filename=exclude_filename))
                        else:
                            self.logger.debug("\t'{exclude_filename}' File Term: {exclude_term}"
                                              .format(exclude_filename=exclude_filename,
                                                      exclude_term=l))
                            exclude_file_keywords.append(l)

        if exclude_file_keywords:
            for unfiltered_file in found_file_list:
                if any(kw in unfiltered_file for kw in exclude_file_keywords):
                    kw_matches = [kw for kw in exclude_file_keywords if kw in unfiltered_file]
                    self.logger.warning("Filtering File. '{exclude_filename}' File Keyword Exclusion(s): "
                                        "{kw} in Path: {path}"
                                        .format(exclude_filename=exclude_filename,
                                                path=unfiltered_file,
                                                kw=kw_matches))
                else:
                    if unfiltered_file not in no_dups_file_list:
                        self.logger.debug('Adding File to Filtered List: {filtered_file}'
                                          .format(filtered_file=unfiltered_file))
                        no_dups_file_list.append(unfiltered_file)
        else:
            for unfiltered_file in found_file_list:
                if unfiltered_file not in no_dups_file_list:
                    self.logger.debug('Adding File to Filtered List: {filtered_file}'
                                      .format(filtered_file=unfiltered_file))
                    no_dups_file_list.append(unfiltered_file)

        if no_dups_file_list:
            self.logger.info('Filtered Files:')
            for f in no_dups_file_list:
                self.logger.info('\t{f}'.format(f=f))
            return no_dups_file_list
        else:
            self.logger.warning('No Files Found while Processing:')
            for path in root_paths:
                self.logger.warning('\t{path}'
                                    .format(path=path))
            return list()

    def __repr__(self):
        return "{self.__class__.__name__}(log_to_console={self.log_to_console})".format(self=self)

    def __str__(self):
        return "{} is a {self.__class__.__name__} object".format(__name__, self=self)

    def process_xml2vhdl(self, **kwargs):
        """Processes project to configure and generate VHDL from XML using ``xml2vhdl``

        Keyword Args:
            **dependencies (list of obj): List of dependency objects in the current project.
            **clean (bool, optional): Cleans the output directories prior to generation.
                Default value: ``False``
            **settings_file (str): Full path of settings file used to generate the current project. Used to
                reference in logging.

        Returns:
            None

        """
        dependencies = funcs.get_kwarg('dependencies', kwargs, None)
        clean = funcs.get_kwarg('clean', kwargs, False)
        settings_file = funcs.get_kwarg('settings_file', kwargs, self.settings_file)

        xml2vhdl_lib_status = dict()
        xml2vhdl_required_libs = list()

        pre_generated_terms = ['pre_generated']
        xml_generated_terms = ['vhd', 'generated']
        vhdl_generated_terms = ['generated']

        for depend in dependencies:
            if depend.name == 'xml2vhdl':
                if isinstance(depend.required_vhdl_libs, (list,)):
                    for lib in depend.required_vhdl_libs:
                        xml2vhdl_lib_status[lib] = False
                        xml2vhdl_required_libs.append(lib)
                else:
                    self.logger.critical('Incorrect setting for XML2VHDL in: {settings}'
                                         .format(settings=settings_file))
                    self.logger.critical('\t"required_vhdl_libs:" entry for "xml2vhdl` Must be a List')
                break

        self.logger.debug('Pre-check XML2VHDL Dependency Status(es):')
        for k, v in xml2vhdl_lib_status.items():
            self.logger.debug('{k}: {v}'.format(k=k, v=v))

        for depend in self.dependencies:
            if depend.name in xml2vhdl_required_libs and depend.vunit_default:
                xml2vhdl_lib_status[depend.name] = True

        self.logger.debug('Post-check XML2VHDL Dependency Status(es):')
        for k, v in xml2vhdl_lib_status.items():
            self.logger.debug('{k}: {v}'.format(k=k, v=v))

        for lib in xml2vhdl_required_libs:
            if not xml2vhdl_lib_status[lib]:
                self.logger.error('XML2VHDL Dependencies Missing: {lib}'.format(lib=lib))
                self.logger.error('Check that it is included, enabled and a unit_default in: {settings}'
                                  .format(settings=settings_file))
                log.errorexit(self.logger)

        self.logger.info('XML2VHDL Support: Enabled')
        self.logger.info('Processing VHDL Dependencies for XML2VHDL Generation...')
        xml2vhdl_settings = xml2xml_iface.Xml2Vhdl(xml2vhdl_path=self.xml2vhdl_path)

        for depend in dependencies:
            if depend.vunit_default:
                # todo: The ``not in`` clause should be redundant once boards and vendor_ip support is
                #       added to :mod:`projectmanager`
                try:
                    if depend.name not in ['boards', 'vendor_ip']:
                        xml2vhdl_settings = self.process_xml2vdl_src_code(xml2vhdl_obj=xml2vhdl_settings,
                                                                          dependency=depend,
                                                                          src_root=os.path.join(depend.src_dir,
                                                                                                'xml'),
                                                                          generated_terms=xml_generated_terms,
                                                                          pre_generated_terms=pre_generated_terms,
                                                                          clean=clean)
                        xml2vhdl_settings = self.process_xml2vdl_src_code(xml2vhdl_obj=xml2vhdl_settings,
                                                                          dependency=depend,
                                                                          src_root=os.path.join(depend.src_dir,
                                                                                                'vhdl'),
                                                                          generated_terms=vhdl_generated_terms,
                                                                          pre_generated_terms=pre_generated_terms,
                                                                          clean=clean)
                except AttributeError as e:
                    self.logger.critical('Could Not Process XML2VHDL for: {depend.name}'
                                         .format(depend=depend))
                    self.logger.critical('{e}'.format(e=e))

        xml2vhdl_settings.generate_vhdl(working_dir=self.xml2vhdl_path)

    def get_top_level_files(self, **kwargs):
        """Gets top-level files from a dictionary of library/absolute file paths.

        Keyword Args:
            **vunit_hdl_dict (dict): Dictionary of list of HDL files, grouped by keys representing the
                corresponding VHDL library.
            **top_level_lib (str): Name of top-level library.
            **top_level (list of str, or str, optional): Top-level filenames.
                Does *not* require file extension. Will search for possible top-levels if not included.
                Default value: ``''``
            **top_ids (list of str, optional): List of terms to determine if HDL top-level filename contains a
                top-level entity. Default value: ``['top_', 'top_']``
            **settings_file (str, optional): Path to settings file to reference in logs.
                Default value: ``''``

        Returns:
            list of str: Absolute paths of top-level files, performs a sys.exit if no top-level files found.

        """
        vunit_hdl_dict = funcs.get_kwarg('vunit_hdl_dict', kwargs, dict())
        top_level_lib = funcs.get_kwarg('top_level_lib', kwargs, '')
        top_level = funcs.get_kwarg('top_level', kwargs, '')
        top_ids = funcs.get_kwarg('top_ids', kwargs, ['_top', '_top'])
        settings_file = funcs.get_kwarg('settings_file', kwargs, '')

        self.logger.info('Finding Top-Level File(s) for: {top_level_lib}'
                         .format(top_level_lib=top_level_lib))

        # \todo This resolves issues where wrapper files are added to None key. This should be stopped
        #       from happening before the script gets here.
        for k, v in vunit_hdl_dict.items():
            self.logger.debug('vunit_hdl_dict: {k}: {v}'.format(k=k, v=v))
            if k is None:
                vunit_hdl_dict.pop(None, None)

        top_level_libs = [key for key in vunit_hdl_dict if top_level_lib in key]
        top_level_files = list()

        if not top_level_libs:
            self.logger.critical('No Top-Level Libraries Found.')
        elif isinstance(top_level, (list,)) and top_level:
            self.logger.debug('Top-Level Files Defined in Settings File: {settings}'
                              .format(settings=settings_file))
            for top in top_level:
                self.logger.debug('\t{top}'
                                  .format(top=top))
                for lib in top_level_libs:
                    for f in vunit_hdl_dict[lib]:
                        filename = os.path.split(f)[1]
                        if top in filename:
                            self.logger.info('Top-Level File: {f}'
                                             .format(f=f))
                            top_level_files.append(f)
        elif isinstance(top_level, (str,)) and top_level != "":
            self.logger.debug('Top-Level Files Defined in Settings File: {settings}'
                              .format(settings=settings_file))
            self.logger.debug('\t{top}'.format(top=top_level))
            for lib in top_level_libs:
                for f in vunit_hdl_dict[lib]:
                    filename = os.path.split(f)[1]
                    if top_level in filename:
                        self.logger.info('Top-Level File: {f}'
                                         .format(f=f))
                        top_level_files.append(f)
        else:
            self.logger.warning('No Top-Level File(s) Defined in Settings File: {settings}'
                                .format(settings=settings_file))

            for lib in top_level_libs:
                for f in vunit_hdl_dict[lib]:
                    filename = os.path.split(f)[1]
                    if top_level_lib in filename and any(t in filename for t in top_ids):
                        self.logger.info('Possible Top-Level File: {f}'
                                         .format(f=f))
                        top_level_files.append(f)

        if top_level_libs and top_level_files:
            self.logger.debug('Top-Level Files:')
            for top in top_level_files:
                self.logger.debug('\t{top}'.format(top=top))

            return top_level_files
        else:
            self.logger.error('No Top-Level Files Found.')
            log.errorexit(self.logger)

    def file_type_detect(self, f_list, ext='.xml'):
        """Detects files of type in list using file extension.

        Args:
            f_list (list): List of file names
            ext (str, optional): File extension to detect. *Must* include ``.``. Default value: ``.xml``

        Returns:
             bool: ``True`` if one file in list is detected, otherwise ``False``

        """
        if isinstance(f_list, (list, )):
            for f in f_list:
                if os.path.splitext(f)[-1] == ext:
                    return True

        return False

    def process_xml2vdl_src_code(self, **kwargs):
        """process source code required by ``xml2vhdl`` generation.

        Keyword Args:
            **xml2vhdl_obj (obj): A :mod:`xml2vhdl_iface.Xml2Vhdl` object to operate on. If one isn't supplied
                one will be generated.
            **dependency (obj): A :mod:`projectmanager.ProjectDependency` object for dependency to process
                for valid``xml2vhdl`` source code.
            **src_root (str): Path to source code location to process.
            **generated_terms (list of str): A list of strings which could be folders where generated
                source code is stored.
            **pre_generated_terms (list of str): A list of strings which could be folders where pre_generated
                source code is stored.
            **clean (bool, optional): Cleans the generated locations using ``shutil.rmtree`` to delete the
                directory(ies) and ``os.makedirs`` to recreate the directory(ies). Default value: ``False``.

        Returns:
            obj: a :mod:`projectflow.Xml2Vhdl` object with attributes updated/completed.

        """
        xml2vhdl_obj = funcs.get_kwarg('xml2vhdl_obj', kwargs, xml2xml_iface.Xml2Vhdl())
        depend = funcs.get_kwarg('dependency', kwargs, None)
        src_root = funcs.get_kwarg('src_root', kwargs, None)
        generated_terms = funcs.get_kwarg('generated_terms', kwargs, list())
        pre_generated_terms = funcs.get_kwarg('pre_generated_terms', kwargs, list())
        clean = funcs.get_kwarg('clean', kwargs, False)

        existing_generated_paths = list()

        if os.path.exists(src_root):
            for root, subpaths, filenames in os.walk(src_root):
                if filenames:
                    self.logger.info('{depend.name}: Found Source(s):'
                                     .format(depend=depend))
                    self.logger.info('\tXML2VHDL Root: {path}'.format(path=src_root))
                    if self.file_type_detect(filenames, '.xml'):
                        xml2vhdl_obj.add_to_path(attr='input_folder', add_paths=src_root)
                        self.logger.info('\tFound: {filenames}'.format(filenames=filenames))
                else:
                    self.logger.info('{depend.name}: No Sources Found for XML2VHDL Generation.'
                                     .format(depend=depend))

                if subpaths:
                    self.logger.debug('{depend.name}: subpaths: {subpaths}'
                                      .format(depend=depend, subpaths=subpaths))
                    for path in subpaths:
                        for term in generated_terms:
                            if term in path:
                                existing_generated_paths.append(os.path.join(src_root, path))
                        for term in pre_generated_terms:
                            if term in path:
                                xml2vhdl_obj.add_to_path(attr='path',
                                                              add_paths=os.path.join(src_root,
                                                                                     path))

                            # Remove pre_generated terms from existing_generated_paths
                            existing_generated_paths = [x for x in existing_generated_paths if term not in x]
                break
        else:
            self.logger.info('{depend.name}: No Sources Found for XML2VHDL Generation.'
                             .format(depend=depend))

        if existing_generated_paths:
            self.logger.info('Found the Following Existing Locations for XML2VHDL Generated Source Code:')
            for path in existing_generated_paths:
                self.logger.info('\t{path}'.format(path=path))
                if clean:
                    self.logger.warning('Cleaning Directory: {path}'.format(path=path))
                    shutil.rmtree(path)
                    os.makedirs(path)

        return xml2vhdl_obj

    def add_vendor_to_exclude_terms(self, exclude_terms=None, **kwargs):
        """Add FPGA vendor to exclude terms.

        Adds `other` FPGA vendors to exclude terms list.

        Args:
            exclude_terms (str or list of str, optional): exclude term(s) to append.

        Keyword Args:
            **vendor (str, optional): FPGA vendor. Valid values: from ``supported_vendors``.
                Default value: ``'xilinx'``
            **supported_vendors (list of str): List of supported FPGA vendors.
                Default value: ``list(['xilinx', 'altera'])``
        Returns:
            list of str: List of exclude terms, with all 'other` FPGA vendors, other than ``vendor``
            argument appended.

        """
        default_vendor_list = ['xilinx', 'altera']
        vendor = funcs.get_kwarg('vendor', kwargs, 'xilinx')
        supported_vendors = funcs.get_kwarg('supported_vendors', kwargs, default_vendor_list)

        if not isinstance(supported_vendors, (list, )):
            self.logger.critical("'supported_vendors' Not a List. Using: {default_vendor_list}"
                                 .format(default_vendor_list=default_vendor_list))

        if vendor not in supported_vendors:
            self.logger.critical("'vendor' Not in Supported Vendor List. Using: {default_vendor}"
                                 .format(default_vendor=default_vendor_list[0]))
            vendor = default_vendor_list[0]

        self.logger.debug('Vendor:            {vendor}'
                          .format(vendor=vendor))
        self.logger.debug('Supported Vendors: {supported_vendors}'
                          .format(supported_vendors=supported_vendors))

        excluded_vendors = [term for term in supported_vendors if term != vendor]
        self.logger.debug('Excluded Vendors:  {excluded_vendors}'
                          .format(excluded_vendors=excluded_vendors))

        if isinstance(exclude_terms, (list, )):
            updated_exclude_terms = exclude_terms
            updated_exclude_terms.extend(excluded_vendors)
        elif isinstance(exclude_terms, (str, )) and exclude_terms != "":
            updated_exclude_terms = [exclude_terms]
            updated_exclude_terms.extend(excluded_vendors)
        else:
            updated_exclude_terms = excluded_vendors

        self.logger.debug('Excluded Terms:    {updated_exclude_terms}'
                          .format(updated_exclude_terms=updated_exclude_terms))

        if not updated_exclude_terms:
            return list()
        else:
            return updated_exclude_terms

    def generate_project_name(self, **kwargs):
        """Generates FPGA Vendor ``project_name`` based on top-level entity name, board name and name override

        If the top-level name and the board name match the board name will be used.
        Keyword Args:
            **top_level_name (str): Top-level entity name
            **project_name_override (str, optional): Project name override
            **board_name (str): Board name

        Returns:
            str: project name if successful, ``None`` if unsuccessful.

        """
        top_level_name = funcs.get_kwarg('top_level_name', kwargs, None)
        project_name_override = funcs.get_kwarg('project_name_override', kwargs, None)
        board_name = funcs.get_kwarg('board_name', kwargs, None)

        if isinstance(project_name_override, (str, )) and project_name_override != "":
            self.logger.info('Using Project Name Override Value as Project Name: {project_name}'
                             .format(project_name=project_name_override))
            return project_name_override

        if board_name is None:
            if isinstance(top_level_name, (str, )) and top_level_name != "":
                self.logger.info('Using Top-Level Entity Name as Project Name: {project_name}'
                                 .format(project_name=top_level_name))
                return top_level_name
        elif isinstance(self.board, (str,)) and self.board != "":
            if isinstance(top_level_name, (str,)) and top_level_name != "":
                if board_name == top_level_name:
                    project_name = board_name
                else:
                    project_name = board_name + '_' + top_level_name
                    self.logger.info('Using Board and Top-Level Entity Name as Project Name: {project_name}'
                                     .format(project_name=project_name))
                return project_name

        self.logger.critical('Could Not Determine Project Name.')
        self.logger.critical('\tTop-Level Name       : {top_level_name'
                             .format(top_level_name=top_level_name))
        self.logger.critical('\tProject Name Override: {project_name_override}'
                             .format(project_name_override=project_name_override))
        self.logger.critical('\tBoard Name           : {board_name}'
                             .format(board_name=board_name))
        return None


if __name__ == '__main__':
    logger.info("projectflow {} {}"
                .format(__version__, version.revision))
    log.sect_break(logger)

    project = ProjectFlow()

else:
    logger.info(__str__)



# -*- coding: utf8 -*-
"""

**************
``docflow.py``
**************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``docflow.py`` is a module provided to setup and manage the auto-documentation generation using ``doxygen``
and ``sphinx``. Project specific values are extracted from the corresponding ``settings.yml`` YAML file
and the projects structure is determined from the ``config.yml`` YAML file, each passed to the module via
:mod:`arguments` (see :ref:`arguments`).

Todo:
    * Consider Moving ``doxygen`` and ``sphinx`` template values to ``config.yml`` YAML file for improved
      flexibility.
    * Determine the name to assign to: ``breathe_project_name``.

"""
import helpers.customlogging as log
import helpers.version as my_version
import projectmanager as project

import helpers.funcs as funcs
import os
import shutil
import subprocess
import datetime
import yaml
import lxml.html

logger = log.config_logger(name=__name__)
LOGLEVEL = log.LogLevelsConsts

rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = my_version.Version(0, 2, 0, svn_rev=rev, disable_svn_logging=True)
__version__ = version.get_version()
__str__ = my_version.about(__name__, __version__, version.revision)


class ProjectDoc(project.RepoFlow):
    """An object describing the configuration and settings for the projects documentation generation.

    Inherits: project.RepoFlow (obj): Inherited RepoFlow object. See :obj:`projectmanager.RepoFlow`.

    Keyword Args:
        **include_dependencies (bool, optional): Process and include project dependencies in the
            generated documentation. Default value: ``False``

    """
    def __init__(self, **kwargs):
        include_dependencies = funcs.get_kwarg('include_dependencies', kwargs, False)
        super(ProjectDoc, self).__init__()

        doxyfiles = list()
        breathe_projects = list()
        self.logger.debug('Configuration:')
        self.logger.debug('\t{self.doxygen_installdir}'
                          .format(self=self))
        self.logger.debug('\t{self.doxypypy_script}'
                          .format(self=self))
        self.logger.debug('\t{self.doxy_template}'
                          .format(self=self))
        self.logger.debug('\t{self.doxy_extra_style_sheets}'
                          .format(self=self))
        self.logger.debug('\t{self.dot_path}'
                          .format(self=self))
        self.logger.debug('\t{self.sphinx_conf_template}'
                          .format(self=self))
        self.logger.debug('\t{self.sphinx_make_template}'
                          .format(self=self))

        try:
            doxy_mainpage_dict = self.doxy_mainpage
        except (AttributeError, KeyError) as e:
            doxy_mainpage_dict = dict()

        try:
            release_uri = self.release_uri
        except (AttributeError, KeyError) as e:
            release_uri = False

        try:
            issue_uri = self.issue_uri
        except (AttributeError, KeyError) as e:
            issue_uri = False

        try:
            releases_document_name = self.releases_document_name
        except (AttributeError, KeyError) as e:
            releases_document_name = False

        try:
            header = self.header
        except AttributeError:
            header = '_static/esdg_header.rst'

        for depend in self.dependencies:
            if depend.top_level:
                self.source_code_publication_msg(name=depend.name,
                                                 restricted_flag=self.restricted_src)
                self.logger.info('Processing Documentation for Top-Level: {name}'.format(name=depend.name))
                doxy_source_path, doxy_build_path = self.set_doxy_paths(doc_root=depend.docs_dir,
                                                                        doxy_version=self.doxy_version,
                                                                        clean=self.doc_clean)
                logo_file = self.set_logo(doc_root=depend.docs_dir,
                                          logo_path=self.logo_path,
                                          logo_name=self.logo_name)

                image_path = self.set_image_path(doc_root=depend.docs_dir,
                                                 image_path=self.image_path)

                source_path = self.set_source_path(doc_root=depend.docs_dir,
                                                   source_path=self.source_path)

                if self.language == 'python':
                    breathe_project = ''
                    src_code_path = os.path.join(depend.scripts_dir, 'python')

                else:
                    src_code_path = depend.src_dir
                    doxy_input_paths = [depend.src_dir, depend.scripts_dir]

                    doxyfile, breathe_project = self.generate_doxyfile(repo_root=self.repo_root,
                                                                       name=depend.name,
                                                                       doxy_input_paths=doxy_input_paths,
                                                                       doxy_source_path=doxy_source_path,
                                                                       doxy_build_path=doxy_build_path,
                                                                       doxy_template=self.doxy_template,
                                                                       doxy_extra_style_sheets=self.doxy_extra_style_sheets,
                                                                       project_logo=logo_file,
                                                                       restricted_src=self.restricted_src,
                                                                       project_title=self.project_title,
                                                                       project_version=self.project_version,
                                                                       doxy_image_path=image_path,
                                                                       project_language=self.language,
                                                                       doxypypy_script=self.doxypypy_script,
                                                                       dot_path=self.dot_path)

                    self.logger.info('Running Doxygen for: {doxy}'.format(doxy=doxyfile))
                    self.run_doxygen(doxyfile=doxyfile)

                doxygen_generated_html_path = os.path.join(doxy_build_path, 'generated', 'html')
                if os.path.exists(doxygen_generated_html_path):
                    self.logger.info('Found Doxygen Generated HTML: {path}'
                                     .format(path=doxygen_generated_html_path))
                    doxy_mainpage_name = self.generate_doxymainpage(doxy_mainpage_dict=doxy_mainpage_dict,
                                                                    doxy_source_path=doxy_source_path,
                                                                    generated_html_path=doxygen_generated_html_path)

                    doxyfile, breathe_project = self.generate_doxyfile(repo_root=self.repo_root,
                                                                       name=depend.name,
                                                                       doxy_input_paths=doxy_input_paths,
                                                                       doxy_source_path=doxy_source_path,
                                                                       doxy_build_path=doxy_build_path,
                                                                       doxy_mainpage_name=doxy_mainpage_name,
                                                                       doxy_template=self.doxy_template,
                                                                       doxy_extra_style_sheets=self.doxy_extra_style_sheets,
                                                                       project_logo=logo_file,
                                                                       restricted_src=self.restricted_src,
                                                                       project_title=self.project_title,
                                                                       project_version=self.project_version,
                                                                       doxy_image_path=image_path,
                                                                       project_language=self.language,
                                                                       doxypypy_script=self.doxypypy_script,
                                                                       dot_path=self.dot_path)

                    self.logger.info('Running Doxygen for: {doxy}'.format(doxy=doxyfile))
                    self.run_doxygen(doxyfile=doxyfile)
                else:
                    self.logger.info('Could Not Find Doxygen Generated HTML: {path}'
                                     .format(path=doxygen_generated_html_path))
                    doxygen_generated_html_path = None

                sphinx_source_path, sphinx_build_path, \
                    sphinx_static_path = self.set_sphinx_paths(doc_root=depend.docs_dir,
                                                               sphinx_version=self.sphinx_version,
                                                               sphinx_static='_static',
                                                               clean=self.doc_clean)

                self.preprocess_sphinxconf(breathe_project=breathe_project,
                                           src_code_path=src_code_path,
                                           sphinx_source_path=sphinx_source_path,
                                           sphinx_build_path=sphinx_build_path,
                                           restricted_src=self.restricted_src,
                                           project_title=self.project_title,
                                           project_version=self.project_version,
                                           project_author=self.project_author,
                                           project_org=self.project_org,
                                           release_uri=release_uri,
                                           issue_uri=issue_uri,
                                           releases_document_name=releases_document_name,
                                           style_sheets=self.doxy_extra_style_sheets,
                                           image_path=image_path,
                                           source_path=source_path,
                                           static_path=sphinx_static_path,
                                           project_logo=logo_file,
                                           toc_depth=4,
                                           sphinx_conf_template=self.sphinx_conf_template,
                                           sphinx_make_template=self.sphinx_make_template,
                                           doxygen_html_path=doxygen_generated_html_path,
                                           project_language=self.language,
                                           header=header)

                self.run_sphinx(source_path=sphinx_source_path,
                                build_path=sphinx_build_path,
                                builder='html',
                                use_make=False)

            elif include_dependencies:
                depend_settings = self.load_settings_from_yaml(settings_root=depend.settings_dir)

                try:
                    doxy_mainpage_dict = depend_settings.doxy_mainpage
                except (AttributeError, KeyError) as e:
                    doxy_mainpage_dict = dict()

                self.source_code_publication_msg(name=depend.name,
                                                 restricted_flag=depend_settings.restricted_src)
                self.logger.info('Processing Documentation for Dependency: {name}'.format(name=depend.name))
                self.set_doxy_paths(doc_root=depend.docs_dir,
                                    doxy_version=self.doxy_version,
                                    clean=depend_settings.doc_clean)

                self.set_logo(doc_root=depend_settings.docs_dir,
                              logo_path=depend_settings.logo_path,
                              logo_name=depend_settings.logo_name)

                image_path = self.set_image_path(doc_root=depend.docs_dir,
                                                 image_path=depend_settings.image_path)

                source_path = self.set_source_path(doc_root=depend.docs_dir,
                                                   source_path=depend_settings.source_path)

                doxy_input_paths = [depend.src_dir, depend.scripts_dir]

                doxyfile, breathe_project = self.generate_doxyfile(repo_root=self.repo_root,
                                                                   name=depend.name,
                                                                   doxy_input_paths=doxy_input_paths,
                                                                   doxy_source_path=doxy_source_path,
                                                                   doxy_build_path=doxy_build_path,
                                                                   doxy_template=self.doxy_template,
                                                                   doxy_extra_style_sheets=self.doxy_extra_style_sheets,
                                                                   project_logo=logo_file,
                                                                   source_path=source_path,
                                                                   restricted_src=depend_settings.restricted_src,
                                                                   project_title=depend_settings.project_title,
                                                                   project_version=depend_settings.project_version,
                                                                   doxy_image_path=image_path,
                                                                   project_language=depend_settings.language,
                                                                   doxypypy_script=self.doxypypy_script,
                                                                   dot_path=self.dot_path)

                self.logger.info('Running Doxygen for: {doxy}'.format(doxy=doxyfile))
                self.run_doxygen(doxyfile=doxyfile)

                doxygen_generated_html_path = os.path.join(doxy_build_path, 'generated', 'html')
                if os.path.exists(doxygen_generated_html_path):
                    self.logger.info('Found Doxygen Generated HTML: {path}'
                                     .format(path=doxygen_generated_html_path))

                    doxy_mainpage_name = self.generate_doxymainpage(doxy_mainpage_dict=doxy_mainpage_dict,
                                                                    doxy_source_path=doxy_source_path,
                                                                    generated_html_path=doxygen_generated_html_path)

                    doxyfile, breathe_project = self.generate_doxyfile(repo_root=self.repo_root,
                                                                       name=depend.name,
                                                                       doxy_input_paths=doxy_input_paths,
                                                                       doxy_source_path=doxy_source_path,
                                                                       doxy_build_path=doxy_build_path,
                                                                       doxy_mainpage_name=doxy_mainpage_name,
                                                                       doxy_template=self.doxy_template,
                                                                       doxy_extra_style_sheets=self.doxy_extra_style_sheets,
                                                                       project_logo=logo_file,
                                                                       restricted_src=self.restricted_src,
                                                                       project_title=self.project_title,
                                                                       project_version=self.project_version,
                                                                       doxy_image_path=image_path,
                                                                       project_language=self.language,
                                                                       doxypypy_script=self.doxypypy_script,
                                                                       dot_path=self.dot_path)

                    self.logger.info('Running Doxygen for: {doxy}'.format(doxy=doxyfile))
                    self.run_doxygen(doxyfile=doxyfile)
                else:
                    self.logger.info('Could Not Find Doxygen Generated HTML: {path}'
                                     .format(path=doxygen_generated_html_path))
                    doxygen_generated_html_path = None

                sphinx_source_path, sphinx_build_path, \
                    sphinx_static_path = self.set_sphinx_paths(doc_root=depend.docs_dir,
                                                               sphinx_version=self.sphinx_version,
                                                               sphinx_static='_static',
                                                               clean=self.doc_clean)

                self.preprocess_sphinxconf(breathe_project=breathe_project,
                                           src_code_path=depend.src_dir,
                                           sphinx_source_path=sphinx_source_path,
                                           sphinx_build_path=sphinx_build_path,
                                           restricted_src=self.restricted_src,
                                           project_title=self.project_title,
                                           project_version=self.project_version,
                                           project_author=self.project_author,
                                           project_org=self.project_org,
                                           release_uri=release_uri,
                                           issue_uri=issue_uri,
                                           releases_document_name=releases_document_name,
                                           style_sheets=self.doxy_extra_style_sheets,
                                           image_path=image_path,
                                           source_path=source_path,
                                           static_path=sphinx_static_path,
                                           project_logo=logo_file,
                                           toc_depth=4,
                                           sphinx_conf_template=self.sphinx_conf_template,
                                           sphinx_make_template=self.sphinx_make_template,
                                           doxygen_html_path=doxygen_generated_html_path,
                                           project_language=self.language,
                                           header=header)

                self.run_sphinx(source_path=sphinx_source_path,
                                build_path=sphinx_build_path,
                                builder='html',
                                use_make=False)

    def __repr__(self):
        return "{self.__class__.__name__}(log_to_console={self.log_to_console})".format(self=self)

    def __str__(self):
        return "{} is a {self.__class__.__name__} object".format(__name__, self=self)

    def source_code_publication_msg(self, **kwargs):
        """ Sends a Critical Warning to the Console if Source Code Will be Published in the Documentation

        Keyword Args:
            **name (str, optional): Name of project which is publishing source code in the
                generated documentation. Default value: ``''``
            **restricted_flag(bool, optional): Flag to determine if source-code is being published in the
                generated documentation. Default value: ``True``

        Returns:
            None

        """
        name = funcs.get_kwarg('name', kwargs, '')
        restricted_flag = funcs.get_kwarg('restricted_flag', kwargs, True)

        if not restricted_flag:
            self.logger.critical('Publishing Source Code in Documentation for: {name}'
                                 .format(name=name))
        else:
            self.logger.info('Source Code NOT Published in Documentation for: {name}'
                             .format(name=name))

    def load_settings_from_yaml(self, **kwargs):
        """Loads a ``settings.yml`` YAML file and returns the 'documentation' key.

        Keyword Args:
            **settings_root (str, optional): Path to the project's root ``settings`` directory.
                Default value: ``''``
            **settings_filename (str, optional): Name of settings file to use.
                Default value: ``default_settings.yml``

        Returns:
            None

        """
        settings_root = funcs.get_kwarg('settings_root', kwargs, '')
        settings_filename = funcs.get_kwarg('settings_filename', kwargs, 'default_settings.yml')

        settings_filename = os.path.join(settings_root, settings_filename)
        if os.path.isfile(settings_filename):
            with open(settings_filename, 'r') as ymlfile:
                settings = yaml.load(ymlfile)
                self.logger.info('Loading Settings from: {settings}'
                                 .format(settings=settings_filename))
        else:
            self.logger.error('Can Not Load Default Settings File: {filename}'
                              .format(filename=settings_filename))
            log.errorexit(self.logger)
        try:
            return settings['documentation']
        except KeyError:
            self.logger.error("Can Not Load 'documentation' Settings from File: {filename}"
                              .format(filename=settings_filename))
            log.errorexit(self.logger)

    def set_logo(self, **kwargs):
        """Sets the full-path to the logo to use in generated documentation.

        Keyword Args:
            **doc_root (str, optional): Path to projects root ``docs`` directory.
                Default value: ``''``
            **logo_path (str, optional): Relative Path, from ``doc_root``, to the project's logo directory.
                Default value: ``''``
            **logo_name (str, optional): Name of logo file to use.
                Default value: ``''``

        Returns:
            str: the full-path to the logo if found on file-system.

            or:
                bool: ``False`` if not found.

        """
        doc_root = funcs.get_kwarg('doc_root', kwargs, '')
        logo_path = funcs.get_kwarg('logo_path', kwargs, '')
        logo_name = funcs.get_kwarg('logo_name', kwargs, '')

        logo = os.path.abspath(os.path.join(doc_root, logo_path, logo_name))
        if os.path.isfile(logo):
            self.logger.info('Using Logo: {logo}'
                             .format(logo=logo))
            return logo
        else:
            self.logger.warning('Logo Missing From File-System: {logo}'
                                .format(logo=logo))
            return False

    def set_image_path(self, **kwargs):
        """Sets the full-path to the images directory to use in generated documentation.

        Keyword Args:
            **doc_root (str, optional): Path to projects root ``docs`` directory.
                Default value: ``''``
            **image_path (str, optional): Relative Path, from ``doc_root``, to the project's image directory.
                Default value: ``''``

        Returns:
            str: the full-path to the images directory if found on file-system.

            or:
                bool: ``False`` if not found.

        """
        doc_root = funcs.get_kwarg('doc_root', kwargs, '')
        image_path = funcs.get_kwarg('image_path', kwargs, '')
        image_path = os.path.abspath(os.path.join(doc_root, image_path))

        if os.path.exists(image_path):
            self.logger.info('Using Image Path: {path}'
                             .format(path=image_path))
            return image_path
        else:
            self.logger.warning('Image Path Missing From File-System: {path}'
                                .format(path=image_path))
            return False

    def set_source_path(self, **kwargs):
        """Sets the full-path to the source directory to use in generated documentation.

        Keyword Args:
            **doc_root (str, optional): Path to projects root ``docs`` directory.
                Default value: ``''``
            **source_path (str, optional): Relative Path, from ``doc_root``, to the project's
                source_path directory. Default value: ``''``

        Returns:
            str: the full-path to the source directory if found on file-system

            or:
                bool: ``False`` if not found.

        """
        doc_root = funcs.get_kwarg('doc_root', kwargs, '')
        source_path = funcs.get_kwarg('source_path', kwargs, '')
        source_path = os.path.abspath(os.path.join(doc_root, source_path))

        if os.path.exists(source_path):
            self.logger.info('Using Source Path: {path}'
                             .format(path=source_path))
            return source_path
        else:
            self.logger.warning('Source Path Missing From File-System: {path}'
                                .format(path=source_path))
            return False

    def set_doxy_paths(self, **kwargs):
        """Sets the ``doxygen`` source and build paths.

        .. note::

           The keyword argument ``doxy_version`` should be available on the host system and this value must be
           included in the :ref:`config_documentation` :ref:`doxygen_config` :ref:`doxygen_supported_tools`
           within the configuration file (see :ref:`configuration`).

        Keyword Args:
            **doc_root (str, optional): Path to projects root ``docs`` directory.
                Default value: ``''``
            **doxy_version (str, optional): version number of ``doxygen`` - determines the build path.
                Default value: ``''``
            **clean (bool, optional): if set to ``True``, removes all existing files from the build directory.
                Default value: ``False``

        Returns:
            tuple: tuple containing:
                * str: Full path of ``doxygen`` source location.
                * str: Full path of ``doxygen`` build location.

        """
        doc_root = funcs.get_kwarg('doc_root', kwargs, '')
        doxy_version = funcs.get_kwarg('doxy_version', kwargs, '')
        clean = funcs.get_kwarg('clean', kwargs, False)

        doxy_source = os.path.join(doc_root, 'doxygen', 'source')
        if not os.path.exists(doxy_source):
            os.makedirs(doxy_source)
            self.logger.info('Creating Doxygen Source Path: {path}'
                             .format(path=doxy_source))
        else:
            self.logger.warning('Doxygen Source Path Exists: {path}'
                                .format(path=doxy_source))

        doxy_build = os.path.join(doc_root, 'doxygen', doxy_version, 'build')
        if not os.path.exists(doxy_build):
            os.makedirs(doxy_build)
            self.logger.info('Creating Doxygen Build Path: {path}'
                             .format(path=doxy_build))
        elif clean:
            self.logger.info('Cleaning Doxygen Build Path: {path}'.
                             format(path=doxy_build))
            shutil.rmtree(doxy_build)
            os.makedirs(doxy_build)
        else:
            self.logger.warning('Doxygen Build Path Exists: {path}'
                                .format(path=doxy_build))

        return doxy_source, doxy_build

    def set_sphinx_paths(self, **kwargs):
        """Sets the ``sphinx`` source, build and static paths.

        .. note::

           The keyword argument ``sphinx_version`` should be available on the host system and this value must
           be included in the :ref:`config_documentation` :ref:`sphinx_config` :ref:`sphinx_supported_tools`
           within the configuration file (see :ref:`configuration`).

        Keyword Args:
            **doc_root (str, optional): Path to projects root ``docs`` directory.
                Default value: ``''``
            **sphinx_version (str, optional): version number of ``sphinx`` to use - determines the build path.
                Default value: ``''``
            **sphinx_static (str, optional): ``sphinx`` static directory name. Where existing source files
                are copied to in order to be included in the generated documentation.
                Default value: ``'_static'``
            **clean (bool, optional): if set to ``True``, removes all existing files from the build directory.
                Default value: ``False``

        Returns:
            tuple: tuple containing:
                * str: Full path of ``sphinx`` source location.
                * str: Full path of ``sphinx`` build location.
                * str: Full path of ``sphinx`` static location.

        """
        doc_root = funcs.get_kwarg('doc_root', kwargs, '')
        sphinx_version = funcs.get_kwarg('sphinx_version', kwargs, '')
        sphinx_static = funcs.get_kwarg('sphinx_static', kwargs, '_static')
        clean = funcs.get_kwarg('clean', kwargs, False)

        sphinx_source = os.path.join(doc_root, 'sphinx', 'source')
        if not os.path.exists(sphinx_source):
            os.makedirs(sphinx_source)
            self.logger.info('Creating Sphinx Source Path: {path}'
                             .format(path=sphinx_source))
        else:
            self.logger.info('Sphinx Source Path Exists: {path}'
                             .format(path=sphinx_source))

        sphinx_static_path = os.path.join(sphinx_source, sphinx_static)
        if not os.path.exists(sphinx_static_path):
            os.makedirs(sphinx_static_path)
            self.logger.info('Creating Sphinx Static Path: {path}'
                             .format(path=sphinx_static_path))
        elif clean:
            self.logger.info('Cleaning Sphinx Static Path: {path}'
                             .format(path=sphinx_static_path))
            shutil.rmtree(sphinx_static_path)
            os.makedirs(sphinx_static_path)
        else:
            self.logger.warning('Sphinx Static Path Exists: {path}'
                                .format(path=sphinx_static_path))

        sphinx_build = os.path.join(doc_root, 'sphinx', sphinx_version, 'build')
        if not os.path.exists(sphinx_build):
            os.makedirs(sphinx_build)
            self.logger.info('Creating Sphinx Build Path: {path}'
                             .format(path=sphinx_build))
        elif clean:
            self.logger.info('Cleaning Sphinx Build Path: {path}'
                             .format(path=sphinx_build))
            shutil.rmtree(sphinx_build)
            os.makedirs(sphinx_build)
        else:
            self.logger.warning('Sphinx Build Path Exists: {path}'
                                .format(path=sphinx_build))

        return sphinx_source, sphinx_build, sphinx_static_path

    def generate_doxymainpage(self, **kwargs):
        """Generates the ``doxygen`` Main Page to act as link between ``doxygen`` and ``sphinx`` documentation

        Parses ``generated_html_path`` looking for HTML files where the filename ends with ``htmlpagefilter``
        and added the file to the generated mainpage, along with the pages title extracted from the HTML page.

        Keyword Args:
            **doxy_mainpage_dict (dict, optional): Dictionary of mainpage settings defining the structure and
                configuration of the mainpage to generate. Default value: ``dict()``
            **doxy_source_path (str, optional): Full path to the ``doxygen`` source path.
                Default value: ``None``
            **generated_html_path (str, optional): Full path to the generated ``doxygen`` generated HTML.
                Default value: ``None``
            **doxy_mainpage_title (str, optional): Tile of the generated mainpage which is referenced in
                the generated documentation.
                Default value: ``'Doxygen Index``
            **htmlpagefilter (str, optional): The suffix of the HTML pages to add to the generated mainpage
                (excluding the file extension).
                Default value: ``'page``

        Returns:
            str: Name of the generated main page

        """
        doxy_mainpage_dict = funcs.get_kwarg('doxy_mainpage_dict', kwargs, dict())
        doxy_source_path = funcs.get_kwarg('doxy_source_path', kwargs, None)
        generated_html_path = funcs.get_kwarg('generated_html_path', kwargs, None)
        doxy_mainpage_title = funcs.get_kwarg('doxy_mainpage_title', kwargs, 'Doxygen Index')
        htmlpagefilter = funcs.get_kwarg('htmlpagefilter', kwargs, 'page')

        try:
            doxy_mainpage_name = doxy_mainpage_dict['mainpage_name']
            self.logger.info('Doxygen Main Page Name: {name}'
                                 .format(name=doxy_mainpage_name))
        except KeyError:
            doxy_mainpage_name = 'README.md'
            self.logger.critical('Doxygen Main Page Name Not Defined defaulting to: {name}'
                                 .format(name=doxy_mainpage_name))

        doxy_mainpage_links_name = os.path.splitext(doxy_mainpage_name)[0].lower() + '_links.md'

        try:
            doxy_autostruct = doxy_mainpage_dict['autostruct']
        except KeyError:
            doxy_autostruct = list()
            self.logger.info('No Additional Content Will Added to Doxygen Main Page when Generating: {name}'
                             .format(name=doxy_mainpage_name))

        try:
            doxy_autogen_pagelinks = doxy_mainpage_dict['autogen_pagelinks']
            if doxy_autogen_pagelinks:
                self.logger.info('Doxygen Main Page Links Will Auto-Generate into: {name}'
                                 .format(name=doxy_mainpage_links_name))
                if doxy_mainpage_links_name not in doxy_autostruct:
                    doxy_autostruct.append(doxy_mainpage_links_name)

                mainpage_links_contents = list()
                htmlfiles = list()
                for dirpath, dirnames, filenames in os.walk(generated_html_path):
                    htmlfiles.extend(filenames)
                    break

                filteredhtmlfiles = [f for f in htmlfiles if os.path.splitext(f)[0].endswith(htmlpagefilter)]
                for pagelink in filteredhtmlfiles:
                    page = lxml.html.parse(os.path.join(generated_html_path, pagelink))
                    title = page.find(".//title").text
                    mainpage_links_contents.append('* \subpage {pagelink} "{title}"'
                                                   .format(pagelink=os.path.splitext(pagelink)[0],
                                                           title=title))

                doxy_mainpage_links_name_path = os.path.join(doxy_source_path, doxy_mainpage_links_name)
                self.logger.info('Generating: {f}'.format(f=doxy_mainpage_links_name_path))
                funcs.writefile_as_list(full_path=doxy_mainpage_links_name_path,
                                        contents=mainpage_links_contents)

        except KeyError:
            doxy_autogen_pagelinks = False
            self.logger.critical('Doxygen Main Page Auto Generate Page Links Not Defined defaulting to: {bool}'
                                 .format(bool=doxy_autogen_pagelinks))

        doxy_mainpage_contents = list()
        doxy_mainpage_contents.append('@mainpage {title}'.format(title=doxy_mainpage_title))
        doxy_mainpage_contents.append('')

        if doxy_autostruct:
            for item in doxy_autostruct:
                item_path = os.path.abspath(os.path.join(doxy_source_path, item))
                if os.path.isfile(item_path):
                    self.logger.info('\tProcessing: {path}'.format(path=item_path))
                    with open(item_path) as mdfile:
                        doxy_mainpage_contents.extend(mdfile.read().splitlines())
                    doxy_mainpage_contents.append('- - -')
                else:
                    self.logger.warning('\tSkipping:   {path}'.format(path=item_path))

        doxy_md_path = os.path.abspath(os.path.join(doxy_source_path, doxy_mainpage_name))
        self.logger.info('Generating: {f}'.format(f=doxy_md_path))
        funcs.writefile_as_list(full_path=doxy_md_path, contents=doxy_mainpage_contents)

        return doxy_mainpage_name

    def generate_doxyfile(self, **kwargs):
        """Generate the ``doxygen`` ``.doxyfile`` from project derived values and template.

        .. warning::

           ``EXAMPLE_PATH`` is set with fixed locations and should be modified to be defined in project
           ``settings.yml`` YAML file.

        Keyword Args:
            **repo_root (str, optional): Full resolved path defined by ``$REPO_ROOT``
                Default value: ``''``
            **name (str, optional): The name of the ``.doxyfile`` to generate.
                Default value: ``''``
            **doxy_input_paths (list of str, optional): List of paths to include as input sources for
                the generation of documentation. Default value: ``list()``
            **doxy_source_path (str, optional): Path to ``doxygen`` source directory which includes
                common documentation files (``*.md``) to include in generated documentation.
                Default value: ``''``
            **doxy_build_path (str, optional): Full path where to build documentation. Default value: ``''``
            **doxy_mainpage_name (str, optional): Name of the mainpage to use as ``index.html`` in
                generated HTML documentation. Default value: ``'README.md'``
            **doxy_template (str, optional): Full path of template file to use for generic ``.doxyfile``
                configuration options. Default value: ``''``
            **doxy_extra_style_sheets (str, optional): Extra CSS style sheets used in the generation of
                HTML pages. Default value: ``''``
            **image_path (str, optional): Full path of images to include in generated documentation.
                Default value: ``False``
            **restricted_src (bool, optional): If set ``True`` stops source code from being published in the
                generated documentation. Default value: ``True``
            **project_title (str, optional): The project title used in the generated documentation.
                Default value: ``''``
            **project_version (str, optional): The project version number, in the form: ``YYYY.NN``
                used in the generated documentation. Default value: ``''``
            **project_logo (bool, optional): Full path to the logo to use in the generated documentation.
                Default value: ``False``
            **project_language (str, optional): The language of the project being generated, used to optimise
                the output of the generated documentation. Default value: ``''``
            **doxypypy_script (str, optional): Full path of ``doxypypy.py`` for parsing ``python`` code in
                ``doxygen``. ``sphinx`` is preferred over this option. Default value: ``''``
            **dot_path (str, optional): Full path to ``dot`` for using ``graphviz`` to generate
                diagrams in generated documentation. Default value: ``''``
        Returns:
            tuple: tuple containing:
                * str: Full path of the generated ``doxyfile``
                * str: Full path of the generated ``xml`` to be used by ``breathe`` to bridge ``doxygen`` and
                  ``sphinx`` generated documentation.

            or:
                int: ``-1`` on failure.

        """
        repo_root = funcs.get_kwarg('repo_root', kwargs, '')
        name = funcs.get_kwarg('name', kwargs, '')
        doxy_input_paths = funcs.get_kwarg('doxy_input_paths', kwargs, list())
        doxy_source_path = funcs.get_kwarg('doxy_source_path', kwargs, '')
        doxy_build_path = funcs.get_kwarg('doxy_build_path', kwargs, '')
        doxy_mainpage_name = funcs.get_kwarg('doxy_mainpage_name', kwargs, 'README.md')
        doxy_template = funcs.get_kwarg('doxy_template', kwargs, '')
        doxy_extra_style_sheets = funcs.get_kwarg('doxy_extra_style_sheets', kwargs, '')
        image_path = funcs.get_kwarg('doxy_image_path', kwargs, False)
        restricted_src = funcs.get_kwarg('restricted_src', kwargs, True)
        project_title = funcs.get_kwarg('project_title', kwargs, '')
        project_version = funcs.get_kwarg('project_version', kwargs, '')
        project_logo = funcs.get_kwarg('project_logo', kwargs, False)
        project_language = funcs.get_kwarg('project_language', kwargs, '')
        doxypypy_script = funcs.get_kwarg('doxypypy_script', kwargs, '')
        dot_path = funcs.get_kwarg('dot_path', kwargs, '')

        if os.path.isfile(doxy_template):
            with open(doxy_template) as doxyfile:
                doxy_template_contents = doxyfile.read()
        else:
            self.logger.critical('Template {template} Does Not Exist on File System'
                                 .format(template=doxy_template))
            return -1

        doxy_settings_header_start = "# Automatically Generated Settings (DO NOT MODIFY)"
        doxy_sectbreak = "#" * 80
        doxy_project_name = "PROJECT_NAME = "
        doxy_project_number = "PROJECT_NUMBER = "
        doxy_project_logo = "PROJECT_LOGO = "
        doxy_project_output_dir = "OUTPUT_DIRECTORY = "
        doxy_strip_from_path = "STRIP_FROM_PATH = "
        doxy_warn_logfile = "WARN_LOGFILE = "
        doxy_md_path = os.path.abspath(os.path.join(doxy_source_path, doxy_mainpage_name))
        if os.path.isfile(doxy_md_path):
            self.logger.info('Using Doxygen MainPage File: {path}'
                             .format(path=doxy_md_path))
            doxy_input = "INPUT = {mainpage} ".format(mainpage=doxy_md_path)
            doxy_use_mdfile_as_mainpage = "USE_MDFILE_AS_MAINPAGE = {mainpage}".format(mainpage=doxy_md_path)
        else:
            self.logger.warning('Cannot Find Doxygen MainPage File: {path}'
                                .format(path=doxy_md_path))
            doxy_input = "INPUT = "
            doxy_use_mdfile_as_mainpage = ""

        doxy_input_list = list()
        doxy_example_path = "EXAMPLE_PATH = "
        doxy_image_path = "IMAGE_PATH = "
        doxy_have_dot = "HAVE_DOT = "
        doxy_dot_path = "DOT_PATH = "
        doxy_html_extra_stylesheet = "HTML_EXTRA_STYLESHEET = "
        doxy_verbatim_headers = "VERBATIM_HEADERS = "
        doxy_source_browser = "SOURCE_BROWSER = "
        doxy_settings_header_end = "# The following Settings are Copied from Template File:"
        doxy_template_location = "# {template}".format(template=doxy_template)

        doxy_input_filter = "INPUT_FILTER = "
        # Language Dependent Settings:
        doxy_opt_output_vhdl = "OPTIMIZE_OUTPUT_VHDL = "
        doxy_opt_output_java = "OPTIMIZE_OUTPUT_JAVA = "
        doxy_filter_patterns = "FILTER_PATTERNS = "

        # Restrict Source Code in Documentation:
        if restricted_src:
            doxy_verbatim_headers += "NO"
            doxy_source_browser += "NO"
        else:
            doxy_verbatim_headers += "YES"
            doxy_source_browser += "YES"

        doxy_project_name = doxy_project_name + '"' + str(project_title) + '"'
        doxy_project_number = doxy_project_number + '"' + str(project_version) + '"'

        # Project Logo
        if project_logo:
            doxy_project_logo = doxy_project_logo + project_logo
        else:
            self.logger.warning('PROJECT_LOGO Not Set')

        # Image Path
        if image_path:
            doxy_image_path += image_path
        else:
            logger.warning('IMAGE_PATH Not Set')

        # and create log and generated folders:
        log_path = os.path.join(doxy_build_path, 'log', 'log.txt')
        output_dir = os.path.join(doxy_build_path, 'generated')
        breathe_project = os.path.join(output_dir, 'xml')
        doxy_project_output_dir = doxy_project_output_dir + output_dir
        doxy_warn_logfile = doxy_warn_logfile + log_path

        for path in doxy_input_paths:
            doxy_input_list.append(path)

        # Strip from Path
        doxy_strip_from_path = doxy_strip_from_path + repo_root

        # Input Paths:
        if project_language == 'python':
            doxy_opt_output_java += "YES"
            doxy_opt_output_vhdl += "NO"
            doxy_filter_patterns = (doxy_filter_patterns + "*.py="
                                    + doxypypy_script)
            doxy_input_list = self._parse_input_list(paths_list=doxy_input_paths,
                                                     term='python')
        elif project_language == 'vhdl':
            doxy_opt_output_java += "NO"
            doxy_opt_output_vhdl += "YES"
            doxy_input_list = self._parse_input_list(paths_list=doxy_input_paths,
                                                     term='vhdl')

        doxy_example_list = self._parse_input_list(paths_list=doxy_input_paths,
                                                   term='html')
        doxy_input_string = ''
        for inc_path in doxy_input_list:
            # Don't Add Examples to Source Code Input List...
            if inc_path not in doxy_example_list:
                doxy_input_string = doxy_input_string + inc_path + ' '
        doxy_input += doxy_input_string

        # Example Path:
        doxy_example_string = ''
        for inc_path in doxy_example_list:
            doxy_example_string = doxy_example_string + inc_path + ' '
        doxy_example_path += doxy_example_string

        # DOT Handling
        doxy_have_dot += "YES"
        doxy_dot_path += dot_path

        # Get Extra Stylesheet(s)
        if isinstance(doxy_extra_style_sheets, (list,)):
            extra_stylesheet = ''
            for style in doxy_extra_style_sheets:
                extra_stylesheet = extra_stylesheet + style + ' '
        else:
            extra_stylesheet = doxy_extra_style_sheets

        doxy_html_extra_stylesheet += extra_stylesheet

        settings_doxyfile = [doxy_sectbreak,
                             doxy_settings_header_start,
                             doxy_sectbreak,
                             doxy_project_name,
                             doxy_project_number,
                             doxy_project_logo,
                             doxy_image_path,
                             doxy_strip_from_path,
                             doxy_input,
                             doxy_example_path,
                             doxy_project_output_dir,
                             doxy_warn_logfile,
                             doxy_html_extra_stylesheet,
                             doxy_have_dot, doxy_dot_path,
                             doxy_opt_output_vhdl,
                             doxy_opt_output_java,
                             doxy_filter_patterns,
                             doxy_input_filter,
                             doxy_verbatim_headers,
                             doxy_source_browser,
                             doxy_use_mdfile_as_mainpage,
                             doxy_sectbreak,
                             doxy_settings_header_end,
                             doxy_template_location,
                             doxy_sectbreak]

        for line in doxy_template_contents.split('\n'):
            settings_doxyfile.append(line)

        doxyfile = os.path.join(doxy_source_path, name + '.doxyfile')
        self.logger.info('Generating: {f}'.format(f=doxyfile))
        funcs.writefile_as_list(full_path=doxyfile, contents=settings_doxyfile)

        return doxyfile, breathe_project

    def _parse_input_list(self, **kwargs):

        paths_list = funcs.get_kwarg('paths_list', kwargs, list())
        term = funcs.get_kwarg('term', kwargs, '')

        parsed_list = list()

        for p in paths_list:
            for root, subdirs, files in os.walk(p):
                for subdir in subdirs:
                    if term in root or term in subdir:
                        parsed_list.append(os.path.join(root, subdir))

        for p in parsed_list:
            self.logger.debug('Found: {path}'
                              .format(path=p))

        return parsed_list

    def run_doxygen(self, **kwargs):
        """``doxygen`` runner

            Keyword Args:
                **doxyfile (str, optional): The full path of ``doxygen`` ``.doxyfile`` used to generate
                    documentation. Default value: ``''``
                **doxy_bin (str, optional): The name of the bin to use to execute ``doxygen``
                    Default value: ``'doxygen'``

            Returns:
                None
            """
        doxyfile = funcs.get_kwarg('doxyfile', kwargs, '')
        doxy_bin = funcs.get_kwarg('doxy_bin', kwargs, 'doxygen')
        subprocess.call([doxy_bin, doxyfile])

    def run_sphinx(self, **kwargs):
        """``sphinx`` runner

        Keyword Args:
            **sphinx_source_path (str, optional): The path of ``sphinx`` source to generate documentation.
                Default value: ``''``
            **sphinx_build_path (str, optional): The path where ``sphinx`` will generate the documentation.
                Default value: ``''``
            **builder (str, optional): builder to invoke.
                Valid values: ``html``. Default value: ``html``
            **use_make (bool ,optional): Use the `Makefile` to generate documentation instead of executing
                directly. Default value: ``False``

        Returns:
            int: ``0`` on Success

        """
        source_path = funcs.get_kwarg('source_path', kwargs, '')
        build_path = funcs.get_kwarg('build_path', kwargs, '')
        builder = funcs.get_kwarg('builder', kwargs, 'html')
        use_make = funcs.get_kwarg('use_make', kwargs, False)

        supported_builders = ['html', 'latex']
        if builder not in supported_builders:
            self.logger.warning('Unsupported Sphinx Builder: {builder}'
                                .format(builder=builder))
            builder = supported_builders[0]
            self.logger.warning('\t Defaulting to: {builder}'
                                .format(builder=builder))

        if use_make:
            self.logger.info('Using Makefile to Build Documentation from Sphinx')
            from_dir = os.getcwd()
            makepath = os.path.abspath(os.path.join(source_path, '..'))
            os.chdir(makepath)
            subprocess.call(['make', builder])
            os.chdir(from_dir)
            return 0

        else:
            subprocess.call(['sphinx-build', '-b', builder, source_path, build_path])
            return 0

    def preprocess_sphinxconf(self, **kwargs):
        """Preprocessing for the ``sphinx`` ``conf.py`` File based on the current settings

        Checks locations of directories and files exist and copies relevant files to documentation source
        directory prior to generating the configuration file.

        Keyword Args:
            **source_path (str, optional): Documentation source path.
                Default value: ``''``

            **src_code_path (str, optional): Top-level path of source code to add to
                ``sphinx`` documentation. This path will searched recursively, directories name: ``template``
                will be added to a list to be used to add template files to the configuration.
                Default value: ``''``
            **sphinx_source_path (str, optional): The path of ``sphinx`` source code to add in the generated
                documentation. Default value: ``''``
            **sphinx_build_path (str, optional): The path where ``sphinx`` will generate the documentation.
                Default value: ``''``
            **project_name (str, optional): The project name. Default value: ``''``
            **project_version (str, optional): The projects documentation version. Default value: ``''``
            **project_release (str, optional): The project release. Default value: ``''``
            **project_author (str, optional): The project documentation's author. Default value: ``''``
            **project_org (str, optional): The project's organisation. Default value: ``''``
            **project_logo (str, optional): The name of the project logo to use in the project's documentation.
                Default value: ``False``
            **release_uri (str, optional): The URL for releases to determine the release version.
                Default value: ``False``
            **issue_uri (str, optional): The URL for releases to determine the issue reference.
                Default value: ``False``
            **releases_document_name (str or list of str, optional): The name of the Changelog used by
                ``releases``. Default value: ``False``
            **style_sheets (str, optional): Additional Style-Sheets to use when generating HTML
                documentation. Default value: ``False``
            **breathe_project (str, optional): Breathe project to use when bridging doxygen generated
                documentation with sphinx. Default value: ``False``
            **toc_depth (int, optional): The depth of the toctree used in the generated documentation.
                Default value: ``4``
            **image_path (str, optional): Path to directory of images to add to the generated
                ``sphinx`` documentation. Files in this directory will be copied to the ``static_path`` used
                by ``sphinx``. Default value: ``False``
            **source_path (str, optional): Path to directory of source files to add to the generated
                ``sphinx`` documentation. Files in this directory will be copied to the ``static_path`` used
                by ``sphinx``. Default value: ``False``
            **static_path (str, optional): Path to directory where source files will be copied for the
                generation of ``sphinx`` documentation. Default value: ``'_static'``
            **doxygen_html_path (str, optional): Path where ``doxygen`` has generated HTML documentation
                prior to generating the ``sphinx`` documentation. These pages will be copied to a 'html'
                directory in the ``sphinx`` ``static_path`` for inclusion in the generated ``sphinx``
                documentation. Default value: ``None'``
            **sphinx_conf_template (str, optional): Relative Path, from ``docflow.py``, to the ``sphinx``
                configuration file template to append to the generated configuration file.
                Default value: ``''``
            **sphinx_make_template (str, optional): Relative Path, from ``docflow.py``, to the ``sphinx``
                ``Makefile`` file template to append to the generated configuration file.
                Default value: ``''``
            **language (str, optional): Language to determine if module index is included.
                Included if set to: ``python``. Excluded if set to anything else. Default value: ``python``
            **header (str, optional): Header file to reference at the top of the generated file.
                Default value: ``_static/esdg_header.rst``

        Returns:
            None
        """
        src_code_path = funcs.get_kwarg('src_code_path', kwargs, '')
        sphinx_source_path = funcs.get_kwarg('sphinx_source_path', kwargs, '')
        sphinx_build_path = funcs.get_kwarg('sphinx_build_path', kwargs, '')
        project_title = funcs.get_kwarg('project_title', kwargs, '')
        project_version = funcs.get_kwarg('project_version', kwargs, '')
        project_author = funcs.get_kwarg('project_author', kwargs, '')
        project_org = funcs.get_kwarg('project_org', kwargs, '')
        project_logo = funcs.get_kwarg('project_logo', kwargs, False)
        release_uri = funcs.get_kwarg('release_uri', kwargs, False)
        issue_uri = funcs.get_kwarg('issue_uri', kwargs, False)
        releases_document_name = funcs.get_kwarg('releases_document_name', kwargs, False)
        style_sheets = funcs.get_kwarg('style_sheets', kwargs, False)
        breathe_project = funcs.get_kwarg('breathe_project', kwargs, '')
        toc_depth = funcs.get_kwarg('toc_depth', kwargs, 4)
        image_path = funcs.get_kwarg('image_path', kwargs, False)
        source_path = funcs.get_kwarg('source_path', kwargs, False)
        static_path = funcs.get_kwarg('static_path', kwargs, os.path.join(sphinx_source_path, '_static'))
        doxygen_html_path = funcs.get_kwarg('doxygen_html_path', kwargs, None)
        sphinx_conf_template = funcs.get_kwarg('sphinx_conf_template', kwargs, '')
        sphinx_make_template = funcs.get_kwarg('sphinx_make_template', kwargs, '')
        project_language = funcs.get_kwarg('project_language', kwargs, 'python')
        header = funcs.get_kwarg('header', kwargs, '_static/esdg_header.rst')

        src_code_path_list = list()
        if os.path.exists(src_code_path):
            self.logger.debug('Source Path Found: {path}'.format(path=src_code_path))
            for root, subpaths, filenames in os.walk(src_code_path):
                if subpaths:
                    for subpath in subpaths:
                        if 'template' in subpath or 'template' in root:
                            self.logger.debug('Skipping template... {path}'
                                              .format(path=os.path.relpath(os.path.join(root, subpath),
                                                                           sphinx_source_path)))
                        else:
                            path_to_add = os.path.relpath(os.path.join(root, subpath), sphinx_source_path)
                            src_code_path_list.append(path_to_add)
            self.logger.info('Found the Following Source Paths (Relative to Sphinx Source Path):')
            for path in src_code_path_list:
                self.logger.info('\t{path}'
                                 .format(path=path))
        else:
            self.logger.critical('Source Path Not Found: {path}'.format(path=src_code_path))

        if image_path:
            self.logger.info('Copying Images to Static Path: {path}'.format(path=static_path))
            funcs.copy_files_from_dir(image_path, static_path)

        if source_path:
            self.logger.info('Copying Sources to Static Path: {path}'.format(path=static_path))
            funcs.copy_files_from_dir(source_path, static_path)

        if project_logo:
            self.logger.info('Copying Logo: {project_logo}'.format(project_logo=project_logo))
            self.logger.info('\tto Static Path: {path}'.format(path=static_path))
            shutil.copy2(project_logo, static_path)

        if style_sheets:
            abs_css_static_path = os.path.abspath(os.path.join(static_path, 'css'))
            css_static_path = os.path.join('_static', 'css')
            if not os.path.exists(abs_css_static_path):
                self.logger.info('Creating Style Sheet Directory: {path}'
                                 .format(path=abs_css_static_path))
                os.makedirs(abs_css_static_path)
            else:
                self.logger.info('Using Existing Style Sheet Directory: {path}'
                                 .format(path=abs_css_static_path))
            css_paths = list()
            if isinstance(style_sheets, (list,)):
                for f in style_sheets:
                    self.logger.info('Copying Style Sheet: {f}'.format(f=f))
                    self.logger.info('\tto Static Path: {path}'.format(path=abs_css_static_path))
                    css_name = os.path.split(f)[1]
                    shutil.copy2(f, os.path.join(abs_css_static_path, css_name))
                    css_paths.append(os.path.join(css_static_path, css_name))
            else:
                self.logger.info('Copying Style Sheet: {style_sheets}'.format(style_sheets=style_sheets))
                self.logger.info('\tto Static Path: {path}'.format(path=abs_css_static_path))
                css_name = os.path.split(style_sheets)[1]
                shutil.copy2(style_sheets, os.path.join(abs_css_static_path, css_name))
                css_paths.append(os.path.join(css_static_path, css_name))

            for css in css_paths:
                self.logger.debug('CSS: {css}'
                                  .format(css=css))

        if not os.path.exists(breathe_project) and self.language != 'python':
            self.logger.critical('Breathe Project Missing: {breathe_project}'
                                 .format(breathe_project=breathe_project))
            self.logger.critical('\tCheck Doxygen Project Generated Successfully...')
            breathe_project = ''

        make_template_contents = funcs.readfile_as_list(full_path=sphinx_make_template)

        try:
            if os.path.exists(doxygen_html_path):
                self.logger.info('Adding Doxygen Generated HTML Files to Sphinx...')
                abs_html_static_path = os.path.abspath(os.path.join(static_path, 'html'))
                html_static_path = os.path.join('_static', 'html')
                if os.path.exists(abs_html_static_path):
                    shutil.rmtree(abs_html_static_path)
                self.logger.info('Copying Doxygen Generated HTML Files to Sphinx...')
                shutil.copytree(doxygen_html_path, abs_html_static_path)
            else:
                self.logger.info('Not Adding Doxygen Generated HTML Files')
        except TypeError:
            pass

        self.generate_sphinx_index_file(source_code_path=src_code_path,
                                        source_path=sphinx_source_path,
                                        src_code_path_list=src_code_path_list,
                                        max_depth=toc_depth,
                                        language=project_language,
                                        header=header)

        self.process_sphinx_conf_file(source_path=sphinx_source_path,
                                      src_code_path_list=src_code_path_list,
                                      project_name=project_title,
                                      project_version=project_version,
                                      project_release='',
                                      style_sheets=css_paths,
                                      project_org=project_org,
                                      project_author=project_author,
                                      project_logo=project_logo,
                                      release_uri=release_uri,
                                      issue_uri=issue_uri,
                                      releases_document_name=releases_document_name,
                                      toc_depth=toc_depth,
                                      static_path=static_path,
                                      breathe_project=breathe_project,
                                      template_file=sphinx_conf_template)

        self.process_sphinx_makefile(build_path=sphinx_build_path,
                                     make_template=make_template_contents)

    def process_sphinx_conf_file(self, **kwargs):
        """Generates ``sphinx`` ``conf.py`` File based on the current settings

        Completes options from passed values and completes the remaining file by copying the contents from
        a template file.

        Keyword Args:
            **source_path (str, optional): Documentation source path.
                Default value: ``''``
            **src_code_path_list (list of str, optional): List of source code paths to parse to add to
                ``sphinx`` documentation. Default value: ``list()``
            **conf_name (str, optional): The name of ``sphinx`` configuration file to generate.
                Default value: ``conf.py``
            **project_name (str, optional): The project name. Default value: ``''``
            **project_version (str, optional): The projects documentation version. Default value: ``''``
            **project_release (str, optional): The project release. Default value: ``''``
            **project_author (str, optional): The project documentation's author. Default value: ``''``
            **project_org (str, optional): The project's organisation. Default value: ``''``
            **project_logo (str, optional): The name of the project logo to use in the project's documentation.
                Default value: ``False``
            **release_uri (str, optional): The URL for releases to determine the release version.
                Default value: ``False``
            **issue_uri (str, optional): The URL for releases to determine the issue reference.
                Default value: ``False``
            **releases_document_name (str or list of str, optional): The name of the Changelog used by
                ``releases``. Default value: ``False``
            **style_sheets (str, optional): Additional Style-Sheets to use when generating HTML
                documentation. Default value: ``False``
            **template_file (str, optional): The name of sphinx configuration template file.
                file extension. Default value: ``''``
            **breathe_project (str, optional): Breathe project to use when bridging doxygen generated
                documentation with sphinx. Default value: ``False``
            **toc_depth (int, optional): The depth of the toctree used in the generated documentation.
                Default value: ``4``
            **indent_size (int, optional): Number of spaces in indent. Default value: ``4``

        Returns:
            None

        """
        source_path = funcs.get_kwarg('source_path', kwargs, '')
        src_code_path_list = funcs.get_kwarg('src_code_path_list', kwargs, list())
        conf_name = funcs.get_kwarg('conf_name', kwargs, 'conf.py')
        project_name = funcs.get_kwarg('project_name', kwargs, '')
        project_version = funcs.get_kwarg('project_version', kwargs, '')
        project_release = funcs.get_kwarg('project_release', kwargs, '')
        project_author = funcs.get_kwarg('project_author', kwargs, '')
        project_org = funcs.get_kwarg('project_org', kwargs, '')
        project_logo = funcs.get_kwarg('project_logo', kwargs, False)
        release_uri = funcs.get_kwarg('release_uri', kwargs, False)
        issue_uri = funcs.get_kwarg('issue_uri', kwargs, False)
        releases_document_name = funcs.get_kwarg('releases_document_name', kwargs, False)
        style_sheets = funcs.get_kwarg('style_sheets', kwargs, False)
        template_file = funcs.get_kwarg('template_file', kwargs, '')
        breathe_project = funcs.get_kwarg('breathe_project', kwargs, False)
        toc_depth = funcs.get_kwarg('toc_depth', kwargs, 4)
        static_path = funcs.get_kwarg('static_path', kwargs, '_static')
        indent_size = funcs.get_kwarg('indent_size', kwargs, 4)

        indent = '{indent}'.format(indent=' ' * indent_size)

        now = datetime.datetime.now()
        year = now.year

        breathe_project_name = 'mybreatheproject'

        conf_contents = list()
        conf_contents.append("# -*- coding: utf-8 -*-")
        conf_contents.append("#" * 80)
        conf_contents.append("# Automatically Generated Settings (DO NOT MODIFY)")
        conf_contents.append("#" * 80)

        conf_contents.append("import os")
        conf_contents.append("import sys")
        conf_contents.append("")

        for path in src_code_path_list:
            conf_contents.append("sys.path.append(os.path.abspath('{path}'))".format(path=path))
        conf_contents.append("")

        conf_contents.append("project = u'{project_name}'".format(project_name=project_name))
        conf_contents.append("copyright = u'{year}, {org}'".format(year=year, org=project_org))
        conf_contents.append("author = u'{author}'".format(author=project_author))
        conf_contents.append("version = u'{project_version}'".format(project_version=project_version))
        conf_contents.append("release = u'{project_release}'".format(project_release=project_release))
        conf_contents.append("master_doc = 'index'")
        conf_contents.append("htmlhelp_basename = '{project_name}'"
                             .format(project_name=project_name.replace(' ', '') + 'doc'))
        conf_contents.append("latex_documents = [")
        conf_contents.append("{indent}(master_doc, '{project_name_replaced}.tex', u'{project_name} Documentation',"
                             .format(indent=indent, project_name_replaced=project_name.replace(' ', ''), project_name=project_name))
        conf_contents.append("{indent} u'{author}', 'manual'),".format(indent=indent, author=project_author))
        conf_contents.append("]")
        conf_contents.append("man_pages = [")
        conf_contents.append("{indent}(master_doc, '{project_name_lower}', u'{project_name} Documentation',"
                             .format(indent=indent, project_name_lower=project_name.lower(), project_name=project_name))
        conf_contents.append("{indent} [author], 1)".format(indent=indent))
        conf_contents.append("]")
        conf_contents.append("texinfo_documents = [")
        conf_contents.append("{indent}(master_doc, '{project_name}', u'{project_name} Documentation',"
                             .format(indent=indent, project_name=project_name))
        conf_contents.append("{indent} author, '{project_name}', '',"
                             .format(indent=indent, project_name=project_name))
        conf_contents.append("{indent} 'Miscellaneous'),".format(indent=indent))
        conf_contents.append("]")
        conf_contents.append("html_theme_options = {")
        conf_contents.append("{indent}'logo_only': False,".format(indent=indent))
        conf_contents.append("{indent}'navigation_depth': {toc_depth},"
                             .format(indent=indent, toc_depth=toc_depth)),
        conf_contents.append("{indent}'collapse_navigation': False,".format(indent=indent))
        conf_contents.append("}")

        if project_logo:
            logo_file = os.path.split(project_logo)[1]
            logo_path = os.path.join(static_path, logo_file)
            if logo_path.startswith(source_path):
                rel_logo_path = logo_path[len(source_path)+1:]
                self.logger.debug('Converted Logo Path to Relative Path: {path}'.format(path=rel_logo_path))
                conf_contents.append("html_logo = '{logo}'".format(logo=rel_logo_path))
            else:
                self.logger.debug('Using Absolute Logo Path: {path}'.format(path=logo_path))
                conf_contents.append("html_logo = '{logo}'".format(logo=logo_path))

        if static_path.startswith(source_path):
            rel_static_path = static_path[len(source_path)+1:]
            self.logger.debug('Converted Static Path to Relative Path: {path}'.format(path=rel_static_path))
            conf_contents.append("html_static_path = ['{static_path}']".format(static_path=rel_static_path))
        else:
            self.logger.debug('Using Absolute Static Path: {path}'.format(path=static_path))
            conf_contents.append("html_static_path = ['{static_path}']".format(static_path=static_path))

        if breathe_project:
            conf_contents.append("breathe_projects = {{'{project}': '{xml_path}'}}"
                                 .format(project=breathe_project_name,
                                         xml_path=breathe_project))
            conf_contents.append("breathe_default_project = '{project}'"
                                 .format(project=breathe_project_name))

        if style_sheets:
            conf_contents.append("html_context = {")
            conf_contents.append("{indent}'css_files': {css_files},"
                                 .format(indent=indent, css_files=style_sheets))
            conf_contents.append("}")

        conf_contents.append("")
        conf_contents.append("releases_unstable_prehistory = True")
        if release_uri:
            if release_uri.endswith('/'):
                release_uri = release_uri + "%s"
            else:
                release_uri = release_uri + "/%s"
            self.logger.info('Using Releases URL: {release_uri}'
                             .format(release_uri=release_uri))
            conf_contents.append("releases_release_uri = '{release_uri}'"
                                 .format(release_uri=release_uri))

        if issue_uri:
            if issue_uri.endswith('/'):
                issue_uri = issue_uri + "%s"
            else:
                issue_uri = issue_uri + "/%s"
            self.logger.info('Using Issue URL: {issue_uri}'
                             .format(issue_uri=issue_uri))
            self.logger.critical('Releases Issue Not Implemented Yet...: {issue_uri}'
                                 .format(issue_uri=issue_uri))
            # conf_contents.append("releases_issue_uri = '{issue_uri}'".format(issue_uri=issue_uri))

        if releases_document_name:
            conf_contents.append("releases_document_name = '{releases_document_name}'"
                                 .format(releases_document_name=releases_document_name))
        conf_contents.append("")
        conf_contents.append("#" * 80)
        conf_contents.append("# The following Settings are Copied from Template File:")
        conf_contents.append("# {template}".format(template=template_file))
        conf_contents.append("#" * 80)

        conf_template_contents = funcs.readfile_as_list(full_path=template_file)
        for line in conf_template_contents:
            conf_contents.append(line)

        conf_file = os.path.join(source_path, conf_name)
        self.logger.info('Generating: {f}'.format(f=conf_file))
        funcs.writefile_as_list(full_path=conf_file, contents=conf_contents)

    def generate_sphinx_modules(self, ** kwargs):
        """Generates ``sphinx`` reStructuredText Module File

        Searches for ``.py`` files in locations found in ``src_code_path_list``, then passes that list to:
        :func:`docflow.ProjectDoc.generate_automodule`

        Keyword Args:
            **source_code_path (str, optional): Top-Level Path of Source-Code Location.
                Default value: ``''``
            **source_path (str, optional): Documentation source path where the ``<filename>.rst`` should be
                generated. Default value: ``''``
            **filename (str, optional): The name of module file to generate. Default value: ``modules.rst``
            **src_code_path_list (list of str, optional): List of source code paths to parse to add to
                Sphinx documentation. Default value: ``list()``
            **indent_size (int, optional): Number of spaces in indent. Default value: ``3``
            **header (str, optional): Header file to reference at the top of the generated file.
                Default value: ``_static/esdg_header.rst``

        Returns:
            None

        """
        source_code_path = funcs.get_kwarg('source_code_path', kwargs, '')
        source_path = funcs.get_kwarg('source_path', kwargs, '')
        filename = funcs.get_kwarg('filename', kwargs, 'modules.rst')
        src_code_path_list = funcs.get_kwarg('src_code_path_list', kwargs, list())
        indent_size = funcs.get_kwarg('indent_size', kwargs, 3)
        header = funcs.get_kwarg('header', kwargs, '_static/esdg_header.rst')

        fullpath = os.path.join(source_path, filename)
        module_contents = list()
        name = os.path.splitext(filename)[0]
        module_contents.append(".. include:: {header}".format(header=header))
        module_contents.append("")

        module_contents.append(".. _{name}:".format(name=name))
        module_contents.append("")
        title = self.generate_index_heading(heading_text=name.capitalize(),
                                            heading_char='#')

        for line in title:
            module_contents.append(line)
        module_contents.append("")

        if src_code_path_list:
            module_list = list()
            self.logger.info('Parsing Source Code Paths for AutoModule...')
            for path in src_code_path_list:
                automodule_exclude = ['egg-info',
                                      os.path.normpath('/build'),
                                      os.path.normpath('/dist'),
                                      os.path.normpath('/lib')]
                if not any(x in path for x in automodule_exclude):
                    abs_path = os.path.abspath(os.path.join(source_code_path, path))
                    for root, subpaths, filenames in os.walk(abs_path):
                        for f in filenames:
                            if os.path.splitext(f)[1] in ['.py']:
                                if f not in ['__init__.py', ]:  # 'funcs.py']:
                                    module_list.append(os.path.splitext(f)[0])
                        break

            for module in sorted(module_list):
                self.logger.info('\t{module}'
                                 .format(module=module))
                automodule = self.generate_automodule(module=module, indent_size=indent_size)
                for line in automodule:
                    module_contents.append(line)

        self.logger.info('Generating Sphinx Modules File: {f}'
                         .format(f=fullpath))
        funcs.writefile_as_list(full_path=fullpath, contents=module_contents)

    def generate_sphinx_index_file(self, **kwargs):
        """Generates ``sphinx`` reStructuredText Index File

        Searches for ``.rst`` and ``.md`` files in the ``source_path`` and adds them to the generated
        index file.

        If a file matching the ``index_name`` is found it is **not** added to the generated index file.

        If a html directory is found, using the ``/html/`` string, it is added **once** to the generated index
        file.

        Keyword Args:
            **source_code_path (str, optional): Top-Level Path of Source-Code Location.
                Default value: ``''``
            **source_path (str, optional): Documentation source path to parse for valid files to add to index.
                Default value: ``''``
            **static_path (str, optional): The name of the ``sphinx`` Static Path. Default value: ``_static``
            **index_name (str, optional): The name of the index file to generate.
                Default value: ``index.rst``
            **readme_name (str, optional): The name of a valid README file. Could have ``.rst`` or ``.md``
                file extension. Default value: ``README``
            **src_code_path_list (list of str, optional): List of source code paths to parse to add to
                Sphinx documentation. These paths are relative to ``source_path``.
                Default value: ``list()``
            **max_depth (int, optional): The maximum depth of the toctree. Default value: ``2``
            **indent_size (int, optional): Number of spaces in indent. Default value: ``3``
            **language (str, optional): Language to determine if module index is included.
                Included if set to: ``python``. Excluded if set to anything else. Default value: ``python``
            **header (str, optional): Common header file to use in generated documentation.
                Default value: ``'_static/esdg_header.rst'``
        Returns:
            None

        """
        source_code_path = funcs.get_kwarg('source_code_path', kwargs, '')
        source_path = funcs.get_kwarg('source_path', kwargs, '')
        static_path = funcs.get_kwarg('static_path', kwargs, '_static')
        index_name = funcs.get_kwarg('index_name', kwargs, 'index.rst')
        readme_name = funcs.get_kwarg('index_name', kwargs, 'README')
        src_code_path_list = funcs.get_kwarg('src_code_path_list', kwargs, list())
        max_depth = funcs.get_kwarg('max_depth', kwargs, 2)
        indent_size = funcs.get_kwarg('indent_size', kwargs, 3)
        language = funcs.get_kwarg('language', kwargs, 'python')
        header = funcs.get_kwarg('header', kwargs, '_static/esdg_header.rst')
        index_contents = list()

        resolved_src_code_path_list = list()
        for path in src_code_path_list:
            resolved_src_code_path_list.append(os.path.abspath(os.path.join(source_path, path)))

        for path in resolved_src_code_path_list:
            self.logger.debug('Resolved Path: {path}'.format(path=path))

        if language == 'python':
            self.generate_sphinx_modules(source_code_path=source_code_path,
                                         source_path=source_path,
                                         filename='modules.rst',
                                         src_code_path_list=resolved_src_code_path_list,
                                         indent_size=indent_size,
                                         header=header)

        doc_srcs = list()
        self.logger.info('Parsing Sphinx Documentation from: {path}'
                         .format(path=source_path))
        for ext in ['.rst', '.md']:
            self.logger.info('Searching for Documentation Sources with {ext} Extension'
                             .format(ext=ext))
            for root, subdirs, filenames in os.walk(source_path):
                for f in filenames:
                    if ext in f:
                        doc_srcs.append(os.path.join(root, f))

        include_ref = False
        filtered_srcs = list()
        static_html_flag = False
        html_id = '/html/'
        for f in doc_srcs:
            if f.startswith(source_path):
                path = os.path.splitext(f[len(source_path) + 1:])[0]
                if readme_name in path:
                    include_ref = (f[len(source_path) + 1:])
                    self.logger.info('Found README file: {readme}'
                                     .format(readme=include_ref))

                if html_id in path:
                    if not static_html_flag:
                        static_html_flag = True
                        self.logger.info('Found "{html_id}" in {static_path}: {path}'
                                         .format(html_id=html_id, static_path=static_path, path=path))
                else:
                    filtered_srcs.append(path)

        for f in filtered_srcs:
            if os.path.splitext(index_name)[0] in f:
                self.logger.info('Skipping Reference to Self: {index}'
                                 .format(index=index_name))
                filtered_srcs.remove(f)
            elif static_path in f:
                self.logger.info('Skipping Reference to {static_path} for: {f}'
                                 .format(static_path=static_path, f=f))
                filtered_srcs.remove(f)
            else:
                self.logger.debug('Found Documentation Source File: {f}'
                                  .format(f=f))

        if include_ref:
            include = self.generate_include(include_ref=include_ref,
                                            indent_size=indent_size)
            for line in include:
                index_contents.append(line)

        toctree = self.generate_toctree(max_depth=max_depth,
                                        indent_size=indent_size,
                                        srcs=filtered_srcs,
                                        reference_static_html=static_html_flag)

        for line in toctree:
            index_contents.append(line)

        indices_and_tables = self.generate_indices_and_tables(indent_size=indent_size,
                                                              language=language)

        for line in indices_and_tables:
            index_contents.append(line)

        indexfile = os.path.join(source_path, index_name)
        self.logger.info('Generating: {f}'.format(f=indexfile))
        funcs.writefile_as_list(full_path=indexfile, contents=index_contents)

    def generate_include(self, **kwargs):
        """Generates ``sphinx`` reStructuredText include directive

        Generates the include directive, in the form:

        .. code-block:: rest

           .. include:: <include_ref>


        Keyword Args:
            **include_ref (str, optional): The name of the reference to include. Default value: ``''``
            **indent_size (int, optional): Number of spaces in indent. Default value: ``3``

        Returns:
            list of str: A line-by-line list of the complete ``sphinx`` reStructuredText include directive

        """
        include_ref = funcs.get_kwarg('include_ref', kwargs, '')
        indent_size = funcs.get_kwarg('indent_size', kwargs, 3)
        indent = '{indent}'.format(indent=' ' * indent_size)
        include = list()

        include.append('.. include:: {include_ref}'.format(include_ref=include_ref))
        include.append('')

        return include

    def generate_automodule(self, **kwargs):
        """Generates ``sphinx`` reStructuredText auto-module directive

        Generates the auto-module directive, in the form:

        .. code-block:: rest

           .. automodule:: module
              :members:


        Keyword Args:
            **module (str, optional): The name of the module to auto-module. Default value: ``False``
            **indent_size (int, optional): Number of spaces in indent. Default value: ``3``

        Returns:
            list of str: A line-by-line list of the complete ``sphinx`` reStructuredText
            auto-module directive

        """
        module = funcs.get_kwarg('module', kwargs, False)
        indent_size = funcs.get_kwarg('indent_size', kwargs, 3)
        indent = '{indent}'.format(indent=' ' * indent_size)
        automodule = list()

        if module:
            automodule.append('.. automodule:: {module}'.format(module=module))
            automodule.append('{indent}:members:'.format(indent=indent))
            automodule.append('')

        return automodule

    def generate_toctree(self, **kwargs):
        """Generates ``sphinx`` reStructuredText Table-of-Contents Tree

        Generates the TOC Tree, in the form:

        .. code-block:: rest

           ===================
           Index:
           ===================

           ..toctree::
             :maxdepth: 2

             srcs[0]
             srcs[1]
             srcs[n]
             Doxygen Index <_static/html/index.html#://>

        Keyword Args:
            **max_depth (int, optional): The maximum depth of the toctree. Default value: ``2``
            **indent_size (int, optional): Number of spaces in indent. Default value: ``3``
            **srcs (list of str, optional): List of srcs to include in the toctree. Default value: ``list()``
            **reference_static_html (bool, optional): Allows the addition of a link to page external to
                ``sphinx``. Default value: ``False``
            **static_index_text (str, optional): The human readable string for the ``reference_static_html``
                link. Default value: ``Doxygen Index``
            **static_index_link (str, optional): The location of the ``reference_static_html`` link, relative
                to the ``sphinx`` top-level source path. Default value: ``_static/html/index.html``

        Returns:
            list of str: A line-by-line list of the complete ``sphinx`` reStructuredText Table-of-Contents
            Tree

        """
        max_depth = funcs.get_kwarg('max_depth', kwargs, 2)
        indent_size = funcs.get_kwarg('indent_size', kwargs, 3)
        srcs = funcs.get_kwarg('srcs', kwargs, list())
        reference_static_html = funcs.get_kwarg('reference_static_html', kwargs, False)
        static_index_text = funcs.get_kwarg('static_index_text', kwargs, 'Doxygen Index')
        static_index_link = funcs.get_kwarg('static_index_link', kwargs, '_static/html/index.html')

        indent = '{indent}'.format(indent=' ' * indent_size)
        toctree = list()

        title = self.generate_index_heading(heading_text='Index:',
                                            heading_char='=')
        for line in title:
            toctree.append(line)

        toctree.append('.. toctree::')
        toctree.append('{indent}:maxdepth: {depth}'.format(indent=indent, depth=max_depth))
        toctree.append('')

        for src in sorted(srcs):
            toctree.append('{indent}{src}'.format(indent=indent, src=src))

        if reference_static_html:
            self.logger.info('Adding Link to Sphinx Index TOC for Doxygen Generated HTML')
            # Adding #:// due to: https://stackoverflow.com/questions/27979803/external-relative-link-in-sphinx-toctree-directive
            toctree.append('{indent}{static_index_text} <{static_index_link}#://>'
                           .format(indent=indent,
                                   static_index_text=static_index_text,
                                   static_index_link=static_index_link))
            toctree.append('{indent}'.format(indent=indent))
        else:
            toctree.append('{indent}'.format(indent=indent))

        return toctree

    def generate_indices_and_tables(self, **kwargs):
        """Generates ``sphinx`` reStructuredText Indices and Tables List

        Generates the TOC Tree, in the form:

        .. code-block:: rest

           ===================
           Indices and Tables:
           ===================

           * :ref:`genindex`
           * :ref:`modindex`
           * :ref:`search`


        Keyword Args:
            **indent_size (int, optional): Number of spaces in indent. Default value: ``3``
            **language (str, optional): Language to determine if module index is included.
                Included if set to: ``python``. Excluded if set to anything else. Default value: ``python``

        Returns:
            list of str: A line-by-line list of the complete ``sphinx`` reStructuredText Indices and Tables

        """
        indent_size = funcs.get_kwarg('indent_size', kwargs, 3)
        language = funcs.get_kwarg('language', kwargs, 'python')
        indent = '{indent}'.format(indent=' ' * indent_size)
        indices = self.generate_index_heading(heading_text='Indices and Tables:',
                                              heading_char='=')

        indices.append('* :ref:`genindex`')
        if language == 'python':
            indices.append('* :ref:`modindex`')
        indices.append('* :ref:`search`')
        indices.append('')

        return indices

    def generate_index_heading(self, **kwargs):
        """Generates reStructuredText headings matching ``sphinx`` recommended formatting.

        see: `<http://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html>`_

        Keyword Args:
            **heading_text (str, optional): The heading string. Default value: ``False``
            **heading_char (str, optional): A single character representing the reStructuredText
                heading level.  Default value: ``#``

                Valid values: ``#``, ``*``, ``=``, ``-``, ``^``, ``"``

        Returns:
            list of str: A line-by-line list of the complete reStructuredText heading

        """
        heading_text = funcs.get_kwarg('heading_text', kwargs, False)
        heading_char = funcs.get_kwarg('heading_char', kwargs, '#')
        heading_len = len(heading_text)

        heading_list = list()

        if heading_text and heading_len > 0:
            if heading_char in ['#', '*']:
                heading_list.append(heading_char * heading_len)
                heading_list.append(heading_text)
                heading_list.append(heading_char * heading_len)
                heading_list.append('')
            elif heading_char in ['=', '-', '^', '"']:
                heading_list.append(heading_text)
                heading_list.append(heading_char * heading_len)
                heading_list.append('')
            else:
                self.logger.critical('Unsupported Heading Char: {char}'.format(char=heading_char))
                return ''
        else:
            return ''
        
        return heading_list

    def process_sphinx_makefile(self, **kwargs):
        """Processes a ``sphinx`` ``Makefile`` from a supplied template.

        Keyword Args:
            **build_path (str, optional): absolute path of the ``sphinx`` build location.
                Default value: ``''``
            **make_file_template (list of str): line-by-line list of ``Makefile`` template contents.
                Default value: ``''``
            **makefile (str, optional): name of the makefile. Default value: ``Makefile``

        Returns:
            None

        """
        build_path = funcs.get_kwarg('build_path', kwargs, '')
        make_template = funcs.get_kwarg('make_template', kwargs, '')
        makefile = funcs.get_kwarg('makefile', kwargs, 'Makefile')
        make_path = os.path.split(build_path)[0]

        makefile = os.path.join(make_path, makefile)
        self.logger.info('Generating: {f}'.format(f=makefile))

        if os.path.isfile(makefile):
            self.logger.warning('Overwriting: {f}'
                                .format(f=makefile))
        else:
            self.logger.info('Writing: {f}'
                             .format(f=makefile))

        with open(makefile, 'w') as f:
            for line in make_template:
                f.write('{line}\n'.format(line=line))


if __name__ == '__main__':
    logger.info("docflow {} {}"
                .format(__version__, version.revision))
    log.sect_break(logger)

    project = ProjectDoc(include_dependencies=False)

else:
    logger.info(__str__)



# -*- coding: utf8 -*-
"""

********************
``helpers/funcs.py``
********************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``funcs.py`` is a helpers module to provide functions commonly used throughout.

"""
import os
import shutil
import re
import version as my_version
import customlogging as log
logger = log.config_logger(name=__name__)

rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = my_version.Version(0, 1, 0, svn_rev=rev, disable_svn_logging=True)
__version__ = version.get_version()
__str__ = my_version.about(__name__, __version__, version.revision)


class Const(object):
    """Allows attributes to be treated like constants where they can be set one time only.

    Raises:
        ConstError if attribute already exists.

    """
    class ConstError(TypeError):
        pass

    def __setattr__(self, name, value):
        if name in self.__dict__:
            raise self.ConstError, "Cannot Reassign Attribute: {}.{}".format(self.__class__.__name__, name)
        self.__dict__[name] = value


def set_kwargs(obj, kwargs):
    """Sets attributes to specified existing object.


    Args:
        obj (str): name of the object to assign attribute(s).
        kwargs (dict): dictionary of key, values, where key is the attribute and value is the vaule to assign

    Returns:
        None

    """
    if kwargs:
        for k, v in kwargs.iteritems():
            setattr(obj, k, v)


def get_kwarg(arg, kwargs, default):
    """Finds arg in kwargs and returns the value. If arg is not found return the provided default value.

    Args:
        arg (str): Name of argument to extract
        kwargs (dict): Dictionary of keys, values to find and return ``arg`` value from.
        default: The value to assign if ``arg`` is not a key in ``kwargs``

    Returns:
        ``default``

    """
    if arg in kwargs:
        return kwargs[arg]
    else:
        return default


def validate_file(f, touch=False):
    """Checks the given file exists on the file-system. If touch is True, create the file if it is missing.

    Args:
        f (str): Full path, including file name, of the file to validate.
        touch (bool, optional): If set creates an empty file if not found on the file-system.
            Default value: ``False``

    Returns:
        ``True`` on success, ``False`` otherwise.

    """
    basedir = os.path.dirname(f)
    if os.path.exists(f):
        logger.info('Found File: {}'
                    .format(f))
        return True
    else:
        logger.critical('File Missing on System: {}'
                        .format(f))
        if touch:
            logger.info('Creating File: {}'
                        .format(f))
            if not os.path.exists(basedir):
                os.makedirs(basedir)
            with open(f, 'a'):
                os.utime(f, None)
            return True
        else:
            return False


def replace_tag_with_attr_value(obj, tag):
    """Replaces a ``<TAG>`` with a corresponding attribute value.

    Looks for ``<TAG>`` (enclosed in <>) in tag string passed to the function, and if found returns a
    string with the tag replaced by the value as ``obj.<TAG>`` attribute.

    If the attribute exists but is not populated with a value an "empty" string ``''`` is returned.
    This allows all <TAG> options in the layout configuration to be optional.

    Args:
        obj (obj): Object to assign replaced ``<TAG>`` as attribute.
        tag (str): string to check for sub-strings enclosed in ``<>``.

    Returns:
        (str): An empty string if corresponding attribute for the ``<TAG>`` is **not** found; or the
        corresponding attribute if found; or the original ``<TAG>`` if anything else.

    """
    if '<' in tag and '>' in tag:
        attribute_tag = strip_tag(tag)
        obj.logger.debug('Processing {tag}: {attribute_tag}'
                         .format(tag=tag, attribute_tag=attribute_tag))
    else:
        obj.logger.debug('Not a Valid Tag: {tag}'
                         .format(tag=tag))
        return tag

    if hasattr(obj, attribute_tag):
        if not getattr(obj, attribute_tag):
            obj.logger.critical('Attribute Not Found: {attribute_tag}'
                                .format(attribute_tag=attribute_tag))
            return ''
        else:
            obj.logger.debug('Valid Tag {tag}: {attribute_tag}'
                             .format(tag=tag, attribute_tag=attribute_tag))

            resolved_attr = str(getattr(obj, attribute_tag))
            obj.logger.debug('Resolved Attribute from Tag: {resolved_attr}'
                             .format(resolved_attr=resolved_attr))
            return re.sub('(<[^>]+>)', resolved_attr, tag)
    else:
        obj.logger.critical('Attribute Not Found: {attribute_tag}'
                            .format(attribute_tag=attribute_tag))
    return tag


def strip_tag(tag):
    """Removes the ``<`` and ``>`` from tag and returns the lowercase string enclosed.

    Args:
        tag (str): UPPERCASE tag enclosed in ``<`` and ``>``

    Returns:
        (str): lowercase ``TAG`` with ``<`` and ``>`` stripped.

    """
    stripped_tag = (tag[tag.find('<') + 1:tag.find('>')]).lower()
    return stripped_tag


def readfile_as_list(full_path):
    """Takes a full-path to file and returns the contents as a list.

    Args:
        full_path (str): Full path, including file name, of the file to read from.

    Returns:
        (list): Line-by-line list of file contents.

    """
    if os.path.isfile(full_path):
        with open(full_path) as f:
            return f.readlines()
    else:
        logger.critical('File Does Not Exist on File-System: {file}'
                         .format(file=full_path))
        return False

def writefile_as_list(full_path, contents):
    """Takes a list and writes contents line-by-line to file.

    Args:
        full_path (str): Full path, including file name, of the file to write.
        contents (list): Line-by-line list of contents to write.

    Returns:
        None

    """

    if os.path.isfile(full_path):
        logger.warning('Overwriting: {f}'
                       .format(f=full_path))
    else:
        logger.info('Writing: {f}'
                    .format(f=full_path))

    with open(full_path, 'w') as f:
        for line in contents:
            f.write('{line}\n'.format(line=line.rstrip()))


def copy_files_from_dir(src_path, dst_path):
    """Copy files from source directory to destination directory.

    Uses ``shutil.copy2`` to copy files recursively from one directory to another directory.

    Args:
        src_path (str): Full-path of the source directory.
        dst_path (str): Full-path of the destination directory.

    Returns:
        None

    """
    files = list()
    for root, subpaths, filenames in os.walk(src_path):
        for f in filenames:
            files.append(os.path.join(root, f))

    for f in files:
        logger.info('\t{f}'.format(f=f))
        shutil.copy2(f, dst_path)


def clean_path(**kwargs):
    """Cleans given path, by removing and recreating empty path.


    Keyword Args:
        **path (str): Absolute path to clean.

    Returns:
        None

    """
    path = get_kwarg('path', kwargs, False)

    if path and path != "":
        if os.path.exists(path):
            shutil.rmtree(path)
            os.makedirs(path)

